# This module defines template tags for evelinks URLs and eve image URLs
# 
# Many tags will work both with their respective eveonline object
# and their respective eve entity ID
# 
# Example:
# character URL on evewho: {{ my_character|evewho_character_url}}
# character URL on evewho: {{ 1456384556|evewho_character_url}}
# 
# For more examples see examples.html
# 
# To add templatetags for additional providers just add the respective
# template functions and let them call the generic functions

from django import template

from ..models import EveCharacter, EveCorporationInfo, EveAllianceInfo
from ..evelinks import eveimageserver, evewho, dotlan, zkillboard

register = template.Library()

_DEFAULT_IMAGE_SIZE = 32


# generic functions

def _generic_character_url(
    provider: object, 
    obj_prop: str, 
    eve_obj: EveCharacter
) -> str:
    """returns character URL for given provider and object"""
    my_func = getattr(provider, 'character_url')
    if isinstance(eve_obj, EveCharacter):       
        return my_func(getattr(eve_obj, obj_prop))
    
    elif eve_obj is None:
        return ''

    else:
        return my_func(eve_obj)
    

def _generic_corporation_url(
    provider: object, 
    obj_prop: str, 
    eve_obj: object
) -> str:
    """returns corporation URL for given provider and object"""
    my_func = getattr(provider, 'corporation_url')
    if isinstance(eve_obj, (EveCharacter, EveCorporationInfo)):        
        return my_func(getattr(eve_obj, obj_prop))

    elif eve_obj is None:
        return ''
        
    else:
        return my_func(eve_obj)

        
def _generic_alliance_url(
    provider: object, 
    obj_prop: str,
    eve_obj: object
) -> str:
    """returns alliance URL for given provider and object"""
    my_func = getattr(provider, 'alliance_url')
    
    if isinstance(eve_obj, EveCharacter):
        if eve_obj.alliance_id:                                   
            return my_func(getattr(eve_obj, obj_prop))
        else:
            return ''
    
    elif isinstance(eve_obj, EveAllianceInfo):        
        return my_func(getattr(eve_obj, obj_prop))

    elif eve_obj is None:
        return ''
        
    else:
        return my_func(eve_obj)


def _generic_evelinks_url(
    provider: object,
    provider_func: str,
    eve_obj: object
) -> str:
    """returns evelinks URL for given provider, function and object"""
    my_func = getattr(provider, provider_func)
    if eve_obj is None:
        return ''
        
    else:
        return my_func(eve_obj)


# evewho

@register.filter
def evewho_character_url(eve_obj: EveCharacter) -> str:    
    """generates an evewho URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """    
    return _generic_character_url(evewho, 'character_id', eve_obj)
    

@register.filter
def evewho_corporation_url(eve_obj: object) -> str:
    """generates an evewho URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """ 
    return _generic_corporation_url(evewho, 'corporation_id', eve_obj)

        
@register.filter
def evewho_alliance_url(eve_obj: object) -> str:    
    """generates an evewho URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """ 
    return _generic_alliance_url(evewho, 'alliance_id', eve_obj)


# dotlan

@register.filter
def dotlan_corporation_url(eve_obj: object) -> str:
    """generates a dotlan URL for the given object
    Works with allianceauth.eveonline objects and eve entity names
    Returns URL or empty string 
    """ 
    return _generic_corporation_url(dotlan, 'corporation_name', eve_obj)

        
@register.filter
def dotlan_alliance_url(eve_obj: object) -> str:    
    """generates a dotlan URL for the given object
    Works with allianceauth.eveonline objects and eve entity names
    Returns URL or empty string 
    """ 
    return _generic_alliance_url(dotlan, 'alliance_name', eve_obj)


@register.filter
def dotlan_region_url(eve_obj: object) -> str:    
    """generates a dotlan URL for the given object
    Works with eve entity names
    Returns URL or empty string 
    """ 
    return _generic_evelinks_url(dotlan, 'region_url', eve_obj)


@register.filter
def dotlan_solar_system_url(eve_obj: object) -> str:    
    """generates a dotlan URL for the given object
    Works with eve entity names
    Returns URL or empty string 
    """ 
    return _generic_evelinks_url(dotlan, 'solar_system_url', eve_obj)


# zkillboard

@register.filter
def zkillboard_character_url(eve_obj: EveCharacter) -> str:    
    """generates a zkillboard URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """   
    return _generic_character_url(zkillboard, 'character_id', eve_obj)
    

@register.filter
def zkillboard_corporation_url(eve_obj: object) -> str:
    """generates a zkillboard URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """
    return _generic_corporation_url(zkillboard, 'corporation_id', eve_obj)

        
@register.filter
def zkillboard_alliance_url(eve_obj: object) -> str:    
    """generates a zkillboard URL for the given object
    Works with allianceauth.eveonline objects and eve entity IDs
    Returns URL or empty string 
    """
    return _generic_alliance_url(zkillboard, 'alliance_id', eve_obj)


@register.filter
def zkillboard_region_url(eve_obj: object) -> str:    
    """generates a zkillboard URL for the given object
    Works with eve entity IDs
    Returns URL or empty string
    """
    return _generic_evelinks_url(zkillboard, 'region_url', eve_obj)


@register.filter
def zkillboard_solar_system_url(eve_obj: object) -> str:    
    """generates zkillboard URL for the given object
    Works with eve entity IDs
    Returns URL or empty string
    """
    return _generic_evelinks_url(zkillboard, 'solar_system_url', eve_obj)


# image urls

@register.filter
def character_portrait_url(
    eve_obj: object, 
    size: int = _DEFAULT_IMAGE_SIZE
) -> str:
    """generates an image URL for the given object
    Works with EveCharacter objects or character IDs
    Returns URL or empty string
    """
    if isinstance(eve_obj, EveCharacter):        
        return eve_obj.portrait_url(size)

    elif eve_obj is None:
        return ''
        
    else:            
        try:
            return EveCharacter.generic_portrait_url(eve_obj, size)
        except ValueError:
            return ''


@register.filter
def corporation_logo_url(
    eve_obj: object, 
    size: int = _DEFAULT_IMAGE_SIZE
) -> str:
    """generates image URL for the given object
    Works with EveCharacter, EveCorporationInfo objects or corporation IDs
    Returns URL or empty string
    """
    if isinstance(eve_obj, EveCorporationInfo):        
        return eve_obj.logo_url(size)

    elif isinstance(eve_obj, EveCharacter):        
        return eve_obj.corporation_logo_url(size)

    elif eve_obj is None:
        return ''
        
    else:            
        try:
            return EveCorporationInfo.generic_logo_url(eve_obj, size)
        except ValueError:
            return ''


@register.filter
def alliance_logo_url(
    eve_obj: object, 
    size: int = _DEFAULT_IMAGE_SIZE
) -> str:
    """generates image URL for the given object
    Works with EveCharacter, EveAllianceInfo objects or alliance IDs
    Returns URL or empty string
    """
    if isinstance(eve_obj, EveAllianceInfo):        
        return eve_obj.logo_url(size)

    elif isinstance(eve_obj, EveCharacter):        
        return eve_obj.alliance_logo_url(size)

    elif eve_obj is None:
        return ''
        
    else:            
        try:
            return EveAllianceInfo.generic_logo_url(eve_obj, size)
        except ValueError:
            return ''


@register.filter
def type_icon_url(
    type_id: int, 
    size: int = _DEFAULT_IMAGE_SIZE
) -> str:
    """generates a icon image URL for the given type ID    
    Returns URL or empty string
    """
    try:
        return eveimageserver.type_icon_url(type_id, size)
    except ValueError:
        return ''


@register.filter
def type_render_url(
    type_id: int, 
    size: int = _DEFAULT_IMAGE_SIZE
) -> str:
    """generates a render image URL for the given type ID    
    Returns URL or empty string
    """
    try:
        return eveimageserver.type_render_url(type_id, size)
    except ValueError:
        return ''



# Generated by Django 2.2.12 on 2020-05-10 19:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('discord', '0002_service_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='discorduser',
            name='activated',
            field=models.DateTimeField(blank=True, default=None, help_text='Date & time this service account was activated', null=True),
        ),
        migrations.AddField(
            model_name='discorduser',
            name='discriminator',
            field=models.CharField(blank=True, default='', help_text="user's discriminator on Discord", max_length=4),
        ),
        migrations.AddField(
            model_name='discorduser',
            name='username',
            field=models.CharField(blank=True, db_index=True, default='', help_text="user's username on Discord", max_length=32),
        ),
        migrations.AlterField(
            model_name='discorduser',
            name='uid',
            field=models.BigIntegerField(db_index=True, help_text="user's ID on Discord"),
        ),
        migrations.AlterField(
            model_name='discorduser',
            name='user',
            field=models.OneToOneField(help_text='Auth user owning this Discord account', on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='discord', serialize=False, to=settings.AUTH_USER_MODEL),
        ),
    ]



from .utils import clean_setting


DISCORD_APP_ID = clean_setting('DISCORD_APP_ID', '')
DISCORD_APP_SECRET = clean_setting('DISCORD_APP_SECRET', '')
DISCORD_BOT_TOKEN = clean_setting('DISCORD_BOT_TOKEN', '')
DISCORD_CALLBACK_URL = clean_setting('DISCORD_CALLBACK_URL', '')
DISCORD_GUILD_ID = clean_setting('DISCORD_GUILD_ID', '')

# max retries of tasks after an error occurred
DISCORD_TASKS_MAX_RETRIES = clean_setting('DISCORD_TASKS_MAX_RETRIES', 3)

# Pause in seconds until next retry for tasks after the API returned an error
DISCORD_TASKS_RETRY_PAUSE = clean_setting('DISCORD_TASKS_RETRY_PAUSE', 60)

# automatically sync Discord users names to user's main character name when created
DISCORD_SYNC_NAMES = clean_setting('DISCORD_SYNC_NAMES', False)



from django.test import TestCase
from allianceauth.tests.auth_utils import AuthUtils

from ..models import NameFormatConfig


class TestNameFormatConfig(TestCase):

    def test_str(self):
        obj = NameFormatConfig.objects.create(
            service_name='mumble',
            format='{{character_name}}'
        )
        obj.states.add(AuthUtils.get_member_state())
        obj.states.add(AuthUtils.get_guest_state())
        expected = 'mumble: Member, Guest'
        self.assertEqual(str(obj), expected)


from django.contrib import admin

from .models import AuthTS, Teamspeak3User, StateGroup
from ...admin import ServicesUserAdmin


@admin.register(Teamspeak3User)
class Teamspeak3UserAdmin(ServicesUserAdmin):    
    list_display = ServicesUserAdmin.list_display + (        
        'uid',
        'perm_key'        
    )
    search_fields = ServicesUserAdmin.search_fields + ('uid', )
    

@admin.register(AuthTS)
class AuthTSgroupAdmin(admin.ModelAdmin):
    ordering = ('auth_group__name', )
    list_select_related = True  
    
    list_display = ('auth_group', '_ts_group')
    list_filter = ('ts_group', )
    
    fields = ('auth_group', 'ts_group')
    filter_horizontal = ('ts_group',)

    def _ts_group(self, obj):
        return [x for x in obj.ts_group.all().order_by('ts_group_id')]
    
    _ts_group.short_description = 'ts groups'
    #_ts_group.admin_order_field = 'profile__state'


@admin.register(StateGroup)
class StateGroupAdmin(admin.ModelAdmin):
    list_display = ('state', 'ts_group')
    search_fields = ('state__name', 'ts_group__ts_group_name')



from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from allianceauth.services.hooks import NameFormatter
from .manager import Ips4Manager
from .models import Ips4User

import logging

logger = logging.getLogger(__name__)


class Ips4Tasks:
    def __init__(self):
        pass

    @classmethod
    def delete_user(cls, user):
        logging.debug("Attempting to delete IPS4 account for %s" % user)
        if cls.has_account(user) and Ips4Manager.delete_user(user.ips4.id):
            user.ips4.delete()
            logger.info("Successfully deactivated IPS4 for user %s" % user)
            return True
        return False

    @staticmethod
    def has_account(user):
        try:
            return user.ips4.id != ''
        except ObjectDoesNotExist:
            return False

    @staticmethod
    def disable():
        logging.debug("Deleting all IPS4 users")
        Ips4User.objects.all().delete()

    @staticmethod
    def get_username(user):
        from .auth_hooks import Ips4Service
        return NameFormatter(Ips4Service(), user).format_name()



from django.conf import settings

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import User as BaseUser, \
    Permission as BasePermission, Group
from django.db.models import Q, F
from allianceauth.services.hooks import ServicesHook
from django.db.models.signals import pre_save, post_save, pre_delete, \
    post_delete, m2m_changed
from django.db.models.functions import Lower
from django.dispatch import receiver
from django.forms import ModelForm
from django.utils.html import format_html
from django.urls import reverse
from django.utils.text import slugify

from allianceauth.authentication.models import State, get_guest_state,\
    CharacterOwnership, UserProfile, OwnershipRecord
from allianceauth.hooks import get_hooks
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo,\
    EveAllianceInfo
from allianceauth.eveonline.tasks import update_character
from .app_settings import AUTHENTICATION_ADMIN_USERS_MAX_GROUPS, \
    AUTHENTICATION_ADMIN_USERS_MAX_CHARS

if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True    
else:
    _has_auto_groups = False


def make_service_hooks_update_groups_action(service):
    """
    Make a admin action for the given service
    :param service: services.hooks.ServicesHook
    :return: fn to update services groups for the selected users
    """
    def update_service_groups(modeladmin, request, queryset):
        if hasattr(service, 'update_groups_bulk'):
            service.update_groups_bulk(queryset)
        else:
            for user in queryset:  # queryset filtering doesn't work here?
                service.update_groups(user)

    update_service_groups.__name__ = str('update_{}_groups'.format(slugify(service.name)))
    update_service_groups.short_description = "Sync groups for selected {} accounts".format(service.title)
    return update_service_groups


def make_service_hooks_sync_nickname_action(service):
    """
    Make a sync_nickname admin action for the given service
    :param service: services.hooks.ServicesHook
    :return: fn to sync nickname for the selected users
    """
    def sync_nickname(modeladmin, request, queryset):
        if hasattr(service, 'sync_nicknames_bulk'):
            service.sync_nicknames_bulk(queryset)
        else:
            for user in queryset:  # queryset filtering doesn't work here?
                service.sync_nickname(user)

    sync_nickname.__name__ = str('sync_{}_nickname'.format(slugify(service.name)))
    sync_nickname.short_description = "Sync nicknames for selected {} accounts".format(service.title)
    return sync_nickname


class QuerysetModelForm(ModelForm):
    # allows specifying FK querysets through kwarg
    def __init__(self, querysets=None, *args, **kwargs):
        querysets = querysets or {}
        super().__init__(*args, **kwargs)
        for field, qs in querysets.items():
            self.fields[field].queryset = qs


class UserProfileInline(admin.StackedInline):
    model = UserProfile
    readonly_fields = ('state',)
    form = QuerysetModelForm
    verbose_name = ''
    verbose_name_plural = 'Profile'

    def get_formset(self, request, obj=None, **kwargs):
        # main_character field can only show current value or unclaimed alts
        # if superuser, allow selecting from any unclaimed main
        query = Q()
        if obj and obj.profile.main_character:
            query |= Q(pk=obj.profile.main_character_id)
            if request.user.is_superuser:
                query |= Q(userprofile__isnull=True)
            else:
                query |= Q(character_ownership__user=obj)
        qs = EveCharacter.objects.filter(query)
        formset = super().get_formset(request, obj=obj, **kwargs)

        def get_kwargs(self, index):
            return {'querysets': {'main_character': EveCharacter.objects.filter(query)}}
        formset.get_form_kwargs = get_kwargs
        return formset

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


def user_profile_pic(obj):
    """profile pic column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists (requires CSS)
    """
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if user_obj.profile.main_character:
        return format_html(
            '<img src="{}" class="img-circle">',
            user_obj.profile.main_character.portrait_url(size=32)
        )
    else:
        return None
user_profile_pic.short_description = ''


def user_username(obj):
    """user column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """    
    link = reverse(
        'admin:{}_{}_change'.format(
            obj._meta.app_label,
            type(obj).__name__.lower()
        ), 
        args=(obj.pk,)
    )
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if user_obj.profile.main_character:
        return format_html(
            '<strong><a href="{}">{}</a></strong><br>{}',
            link, 
            user_obj.username,
            user_obj.profile.main_character.character_name
        )
    else:
        return format_html(
            '<strong><a href="{}">{}</a></strong>',
            link, 
            user_obj.username,
        )

user_username.short_description = 'user / main'
user_username.admin_order_field = 'username'


def user_main_organization(obj):
    """main organization column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if not user_obj.profile.main_character:
        result = None
    else:        
        corporation = user_obj.profile.main_character.corporation_name
        if user_obj.profile.main_character.alliance_id:        
            result = format_html('{}<br>{}',
                corporation, 
                user_obj.profile.main_character.alliance_name
            )
        else:
            result = corporation    
    return result

user_main_organization.short_description = 'Corporation / Alliance (Main)'
user_main_organization.admin_order_field = \
    'profile__main_character__corporation_name'


class MainCorporationsFilter(admin.SimpleListFilter):
    """Custom filter to filter on corporations from mains only

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    title = 'corporation'
    parameter_name = 'main_corporation_id__exact'

    def lookups(self, request, model_admin):
        qs = EveCharacter.objects\
            .exclude(userprofile=None)\
            .values('corporation_id', 'corporation_name')\
            .distinct()\
            .order_by(Lower('corporation_name'))
        return tuple(
            [(x['corporation_id'], x['corporation_name']) for x in qs]
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        else:    
            if qs.model == User:
                return qs\
                    .filter(profile__main_character__corporation_id=\
                        self.value())
            else:
                return qs\
                    .filter(user__profile__main_character__corporation_id=\
                        self.value())
            

class MainAllianceFilter(admin.SimpleListFilter):
    """Custom filter to filter on alliances from mains only

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    title = 'alliance'
    parameter_name = 'main_alliance_id__exact'

    def lookups(self, request, model_admin):
        qs = EveCharacter.objects\
            .exclude(alliance_id=None)\
            .exclude(userprofile=None)\
            .values('alliance_id', 'alliance_name')\
            .distinct()\
            .order_by(Lower('alliance_name'))
        return tuple(
            [(x['alliance_id'], x['alliance_name']) for x in qs]
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        else:    
            if qs.model == User:
                return qs\
                    .filter(profile__main_character__alliance_id=self.value())                
            else:
                return qs\
                    .filter(user__profile__main_character__alliance_id=\
                        self.value())
                

def update_main_character_model(modeladmin, request, queryset):    
    tasks_count = 0
    for obj in queryset:
        if obj.profile.main_character:
            update_character.delay(obj.profile.main_character.character_id)
            tasks_count += 1

    modeladmin.message_user(
        request, 
        'Update from ESI started for {} characters'.format(tasks_count)
    )

update_main_character_model.short_description = \
    'Update main character model from ESI'


class UserAdmin(BaseUserAdmin):
    """Extending Django's UserAdmin model
    
    Behavior of groups and characters columns can be configured via settings

    """

    class Media:
        css = {
            "all": ("authentication/css/admin.css",)
        }
         
    class RealGroupsFilter(admin.SimpleListFilter):
        """Custom filter to get groups w/o Autogroups"""
        title = 'group'
        parameter_name = 'group_id__exact'

        def lookups(self, request, model_admin):
            qs = Group.objects.all().order_by(Lower('name'))
            if _has_auto_groups:
                qs = qs\
                    .filter(managedalliancegroup__isnull=True)\
                    .filter(managedcorpgroup__isnull=True)                
            return tuple([(x.pk, x.name) for x in qs])

        def queryset(self, request, queryset):
            if self.value() is None:
                return queryset.all()
            else:    
                return queryset.filter(groups__pk=self.value())

    def get_actions(self, request):
        actions = super(BaseUserAdmin, self).get_actions(request)

        actions[update_main_character_model.__name__] = (
            update_main_character_model, 
            update_main_character_model.__name__, 
            update_main_character_model.short_description
        )

        for hook in get_hooks('services_hook'):
            svc = hook()
            # Check update_groups is redefined/overloaded
            if svc.update_groups.__module__ != ServicesHook.update_groups.__module__:
                action = make_service_hooks_update_groups_action(svc)
                actions[action.__name__] = (
                    action, 
                    action.__name__,
                    action.short_description
                )
            
            # Create sync nickname action if service implements it
            if svc.sync_nickname.__module__ != ServicesHook.sync_nickname.__module__:
                action = make_service_hooks_sync_nickname_action(svc)
                actions[action.__name__] = (
                    action, action.__name__, 
                    action.short_description
                )
        return actions

    def _list_2_html_w_tooltips(self, my_items: list, max_items: int) -> str:    
        """converts list of strings into HTML with cutoff and tooltip"""
        items_truncated_str = ', '.join(my_items[:max_items])
        if not my_items:
            result = None
        elif len(my_items) <= max_items:
            result = items_truncated_str
        else:
            items_truncated_str += ', (...)'
            items_all_str = ', '.join(my_items)
            result = format_html(
                '<span data-tooltip="{}" class="tooltip">{}</span>',
                items_all_str,
                items_truncated_str
            )
        return result
    
    inlines = BaseUserAdmin.inlines + [UserProfileInline]

    ordering = ('username', )
    list_select_related = True
    show_full_result_count = True 
    
    list_display = (
        user_profile_pic,
        user_username, 
        '_state', 
        '_groups',
        user_main_organization,
        '_characters',
        'is_active',
        'date_joined',
        '_role'
    )    
    list_display_links = None

    list_filter = ( 
        'profile__state',
        RealGroupsFilter,        
        MainCorporationsFilter,        
        MainAllianceFilter,
        'is_active',
        'date_joined',
        'is_staff',
        'is_superuser'
    )
    search_fields = (
        'username', 
        'character_ownerships__character__character_name'
    )
    
    def _characters(self, obj):
        my_characters = [
            x.character.character_name 
            for x in CharacterOwnership.objects\
                .filter(user=obj)\
                .order_by('character__character_name')\
                .select_related()
        ]
        return self._list_2_html_w_tooltips(
            my_characters, 
            AUTHENTICATION_ADMIN_USERS_MAX_CHARS
        )
          
    _characters.short_description = 'characters'
    

    def _state(self, obj):
        return obj.profile.state.name
    
    _state.short_description = 'state'
    _state.admin_order_field = 'profile__state'

    def _groups(self, obj):
        if not _has_auto_groups:
            my_groups = [x.name for x in obj.groups.order_by('name')]
        else:
            my_groups = [
                x.name for x in obj.groups\
                    .filter(managedalliancegroup__isnull=True)\
                    .filter(managedcorpgroup__isnull=True)\
                    .order_by('name')
            ]
        
        return self._list_2_html_w_tooltips(
            my_groups, 
            AUTHENTICATION_ADMIN_USERS_MAX_GROUPS
        )
       
    _groups.short_description = 'groups'

    def _role(self, obj):
        if obj.is_superuser:
            role = 'Superuser'
        elif obj.is_staff:
            role = 'Staff'
        else:
            role = 'User'        
        return role
    
    _role.short_description = 'role'
    
    def has_change_permission(self, request, obj=None):
        return request.user.has_perm('auth.change_user')

    def has_add_permission(self, request, obj=None):
        return request.user.has_perm('auth.add_user')

    def has_delete_permission(self, request, obj=None):
        return request.user.has_perm('auth.delete_user')

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """overriding this formfield to have sorted lists in the form"""
        if db_field.name == "groups":
            kwargs["queryset"] = Group.objects.all().order_by(Lower('name'))
        return super().formfield_for_manytomany(db_field, request, **kwargs)


@admin.register(State)
class StateAdmin(admin.ModelAdmin):    
    list_select_related = True
    list_display = ('name', 'priority', '_user_count')
    
    def _user_count(self, obj):
        return obj.userprofile_set.all().count()
    _user_count.short_description = 'Users'

    fieldsets = (
        (None, {
            'fields': ('name', 'permissions', 'priority'),
        }),
        ('Membership', {
            'fields': (
                'public', 
                'member_characters', 
                'member_corporations', 
                'member_alliances'
            ),
        })
    )
    filter_horizontal = [
        'member_characters', 
        'member_corporations', 
        'member_alliances', 
        'permissions'
    ]

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """overriding this formfield to have sorted lists in the form"""
        if db_field.name == "member_characters":
            kwargs["queryset"] = EveCharacter.objects.all()\
                .order_by(Lower('character_name'))
        elif db_field.name == "member_corporations":
            kwargs["queryset"] = EveCorporationInfo.objects.all()\
                .order_by(Lower('corporation_name'))
        elif db_field.name == "member_alliances":
            kwargs["queryset"] = EveAllianceInfo.objects.all()\
                .order_by(Lower('alliance_name'))
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def has_delete_permission(self, request, obj=None):
        if obj == get_guest_state():
            return False
        return super(StateAdmin, self).has_delete_permission(request, obj=obj)

    def get_fieldsets(self, request, obj=None):
        if obj == get_guest_state():
            return (
                (None, {
                    'fields': ('permissions', 'priority'),
                }),
            )
        return super(StateAdmin, self).get_fieldsets(request, obj=obj)
    

class BaseOwnershipAdmin(admin.ModelAdmin):
    class Media:
        css = {
            "all": ("authentication/css/admin.css",)
        }
     
    list_select_related = True
    list_display = (
        user_profile_pic,
        user_username,
        user_main_organization,
        'character',
    )
    search_fields = (
        'user__username', 
        'character__character_name', 
        'character__corporation_name', 
        'character__alliance_name'
    )
    list_filter = (                 
        MainCorporationsFilter,        
        MainAllianceFilter,
    )

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.pk:
            return 'owner_hash', 'character'
        return tuple()


@admin.register(OwnershipRecord)
class OwnershipRecordAdmin(BaseOwnershipAdmin):
    list_display = BaseOwnershipAdmin.list_display + ('created',)


@admin.register(CharacterOwnership)
class CharacterOwnershipAdmin(BaseOwnershipAdmin):
    def has_add_permission(self, request):
        return False


class PermissionAdmin(admin.ModelAdmin):
    actions = None
    readonly_fields = [field.name for field in BasePermission._meta.fields]
    list_display = ('admin_name', 'name', 'codename', 'content_type')
    list_filter = ('content_type__app_label',)

    @staticmethod
    def admin_name(obj):
        return str(obj)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_module_permission(self, request):
        return True

    def has_change_permission(self, request, obj=None):
        # can see list but not edit it
        return not obj


# Hack to allow registration of django.contrib.auth models in our authentication app
class User(BaseUser):
    class Meta:
        proxy = True
        verbose_name = BaseUser._meta.verbose_name
        verbose_name_plural = BaseUser._meta.verbose_name_plural


class Permission(BasePermission):
    class Meta:
        proxy = True
        verbose_name = BasePermission._meta.verbose_name
        verbose_name_plural = BasePermission._meta.verbose_name_plural


try:
    admin.site.unregister(BaseUser)
finally:
    admin.site.register(User, UserAdmin)
    admin.site.register(Permission, PermissionAdmin)


@receiver(pre_save, sender=User)
def redirect_pre_save(sender, signal=None, *args, **kwargs):
    pre_save.send(BaseUser, *args, **kwargs)


@receiver(post_save, sender=User)
def redirect_post_save(sender, signal=None, *args, **kwargs):
    post_save.send(BaseUser, *args, **kwargs)


@receiver(pre_delete, sender=User)
def redirect_pre_delete(sender, signal=None, *args, **kwargs):
    pre_delete.send(BaseUser, *args, **kwargs)


@receiver(post_delete, sender=User)
def redirect_post_delete(sender, signal=None, *args, **kwargs):
    post_delete.send(BaseUser, *args, **kwargs)


@receiver(m2m_changed, sender=User.groups.through)
def redirect_m2m_changed_groups(sender, signal=None, *args, **kwargs):
    m2m_changed.send(BaseUser, *args, **kwargs)


@receiver(m2m_changed, sender=User.user_permissions.through)
def redirect_m2m_changed_permissions(sender, signal=None, *args, **kwargs):
    m2m_changed.send(BaseUser, *args, **kwargs)



from unittest.mock import patch, Mock
import urllib

from requests.exceptions import HTTPError

from django.contrib.auth.models import Group, User
from django.test import TestCase

from allianceauth.tests.auth_utils import AuthUtils

from . import (
    TEST_GUILD_ID, 
    TEST_USER_NAME, 
    TEST_USER_ID, 
    TEST_MAIN_NAME, 
    TEST_MAIN_ID, 
    MODULE_PATH,
    ROLE_ALPHA,
    ROLE_BRAVO,
    ROLE_CHARLIE, 
)
from ..discord_client.tests import create_matched_role
from ..app_settings import (
    DISCORD_APP_ID, 
    DISCORD_APP_SECRET,     
    DISCORD_CALLBACK_URL,    
)
from ..discord_client import DiscordClient, DiscordApiBackoff
from ..models import DiscordUser
from ..utils import set_logger_to_file


logger = set_logger_to_file(MODULE_PATH + '.managers', __file__)


@patch(MODULE_PATH + '.managers.DISCORD_GUILD_ID', TEST_GUILD_ID)
@patch(MODULE_PATH + '.managers.DiscordClient', spec=DiscordClient)
@patch(MODULE_PATH + '.models.DiscordUser.objects._exchange_auth_code_for_token')
@patch(MODULE_PATH + '.models.DiscordUser.objects.user_group_names')
@patch(MODULE_PATH + '.models.DiscordUser.objects.user_formatted_nick')
class TestAddUser(TestCase):
        
    def setUp(self):
        self.user = AuthUtils.create_user(TEST_USER_NAME)
        self.user_info = {
            'id': TEST_USER_ID, 
            'name': TEST_USER_NAME,
            'username': TEST_USER_NAME,
            'discriminator': '1234',
        }
        self.access_token = 'accesstoken'
    
    def test_can_create_user_no_roles_no_nick(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = []                
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.return_value = True
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        args, kwargs = mock_DiscordClient.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertIsNone(kwargs['role_ids'])
        self.assertIsNone(kwargs['nick'])

    def test_can_create_user_with_roles_no_nick(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,        
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):
        roles = [
            create_matched_role(ROLE_ALPHA), 
            create_matched_role(ROLE_BRAVO), 
            create_matched_role(ROLE_CHARLIE)
        ]
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = ['a', 'b', 'c']        
        mock_exchange_auth_code_for_token.return_value = self.access_token        
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = roles
        mock_DiscordClient.return_value.add_guild_member.return_value = True
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        args, kwargs = mock_DiscordClient.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertSetEqual(set(kwargs['role_ids']), {1, 2, 3})
        self.assertIsNone(kwargs['nick'])

    @patch(MODULE_PATH + '.managers.DISCORD_SYNC_NAMES', True)
    def test_can_create_user_no_roles_with_nick(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,        
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):        
        mock_user_formatted_nick.return_value = TEST_MAIN_NAME
        mock_user_group_names.return_value = []        
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.return_value = True
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        args, kwargs = mock_DiscordClient.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertIsNone(kwargs['role_ids'])
        self.assertEqual(kwargs['nick'], TEST_MAIN_NAME)

    @patch(MODULE_PATH + '.managers.DISCORD_SYNC_NAMES', False)
    def test_can_create_user_no_roles_and_without_nick_if_turned_off(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):        
        mock_user_formatted_nick.return_value = TEST_MAIN_NAME
        mock_user_group_names.return_value = []        
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.return_value = True
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        args, kwargs = mock_DiscordClient.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertIsNone(kwargs['role_ids'])
        self.assertIsNone(kwargs['nick'])
    
    def test_can_activate_existing_guild_member(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = []                
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.return_value = None
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        
    def test_return_false_when_user_creation_fails(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):        
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = []                
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.return_value = False
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)

    def test_return_false_when_on_api_backoff(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):        
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = []        
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_DiscordClient.return_value.add_guild_member.side_effect = \
            DiscordApiBackoff(999)
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)

    def test_return_false_on_http_error(
        self, 
        mock_user_formatted_nick,
        mock_user_group_names,                
        mock_exchange_auth_code_for_token, 
        mock_DiscordClient
    ):        
        mock_user_formatted_nick.return_value = None
        mock_user_group_names.return_value = []        
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = self.user_info
        mock_DiscordClient.return_value.match_or_create_roles_from_names\
            .return_value = []
        mock_exception = HTTPError('error')
        mock_exception.response = Mock()
        mock_exception.response.status_code = 500
        mock_DiscordClient.return_value.add_guild_member.side_effect = mock_exception
        
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_DiscordClient.return_value.add_guild_member.called)
        

class TestOauthHelpers(TestCase):
        
    @patch(MODULE_PATH + '.managers.DISCORD_APP_ID', '123456')    
    def test_generate_bot_add_url(self):
        bot_add_url = DiscordUser.objects.generate_bot_add_url()

        auth_url = DiscordClient.OAUTH_BASE_URL
        real_bot_add_url = (
            f'{auth_url}?client_id=123456&scope=bot'
            f'&permissions={DiscordUser.objects.BOT_PERMISSIONS}'
        )
        self.assertEqual(bot_add_url, real_bot_add_url)

    def test_generate_oauth_redirect_url(self):
        oauth_url = DiscordUser.objects.generate_oauth_redirect_url()

        self.assertIn(DiscordClient.OAUTH_BASE_URL, oauth_url)
        self.assertIn('+'.join(DiscordUser.objects.SCOPES), oauth_url)
        self.assertIn(DISCORD_APP_ID, oauth_url)
        self.assertIn(urllib.parse.quote_plus(DISCORD_CALLBACK_URL), oauth_url)

    @patch(MODULE_PATH + '.managers.OAuth2Session')
    def test_process_callback_code(self, oauth):
        instance = oauth.return_value
        instance.fetch_token.return_value = {'access_token': 'mywonderfultoken'}

        token = DiscordUser.objects._exchange_auth_code_for_token('12345')

        self.assertTrue(oauth.called)
        args, kwargs = oauth.call_args
        self.assertEqual(args[0], DISCORD_APP_ID)
        self.assertEqual(kwargs['redirect_uri'], DISCORD_CALLBACK_URL)
        self.assertTrue(instance.fetch_token.called)
        args, kwargs = instance.fetch_token.call_args
        self.assertEqual(args[0], DiscordClient.OAUTH_TOKEN_URL)
        self.assertEqual(kwargs['client_secret'], DISCORD_APP_SECRET)
        self.assertEqual(kwargs['code'], '12345')
        self.assertEqual(token, 'mywonderfultoken')


class TestUserFormattedNick(TestCase):
    
    def setUp(self):        
        self.user = AuthUtils.create_user(TEST_USER_NAME)
    
    def test_return_nick_when_user_has_main(self):
        AuthUtils.add_main_character_2(self.user, TEST_MAIN_NAME, TEST_MAIN_ID)
        result = DiscordUser.objects.user_formatted_nick(self.user)
        expected = TEST_MAIN_NAME
        self.assertEqual(result, expected)

    def test_return_none_if_user_has_no_main(self):        
        result = DiscordUser.objects.user_formatted_nick(self.user)        
        self.assertIsNone(result)


class TestUserGroupNames(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.group_1 = Group.objects.create(name='Group 1')
        cls.group_2 = Group.objects.create(name='Group 2')
    
    def setUp(self):
        self.user = AuthUtils.create_member(TEST_USER_NAME)        

    def test_return_groups_and_state_names_for_user(self):
        self.user.groups.add(self.group_1)
        result = DiscordUser.objects.user_group_names(self.user)
        expected = ['Group 1', 'Member']
        self.assertSetEqual(set(result), set(expected))
    
    def test_return_state_only_if_user_has_no_groups(self):
        result = DiscordUser.objects.user_group_names(self.user)
        expected = ['Member']
        self.assertSetEqual(set(result), set(expected))


class TestUserHasAccount(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_user(TEST_USER_NAME)

    def test_return_true_if_user_has_account(self):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        self.assertTrue(DiscordUser.objects.user_has_account(self.user))

    def test_return_false_if_user_has_no_account(self):
        self.assertFalse(DiscordUser.objects.user_has_account(self.user))

    def test_return_false_if_user_does_not_exist(self):        
        my_user = User(username='Dummy')
        self.assertFalse(DiscordUser.objects.user_has_account(my_user))

    def test_return_false_if_not_called_with_user_object(self):                
        self.assertFalse(DiscordUser.objects.user_has_account('abc'))


@patch(MODULE_PATH + '.managers.DiscordClient', spec=DiscordClient)
@patch(MODULE_PATH + '.managers.logger')
class TestServerName(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_user(TEST_USER_NAME)
    
    def test_returns_name_when_api_returns_it(self, mock_logger, mock_DiscordClient):
        server_name = "El Dorado"
        mock_DiscordClient.return_value.guild_name.return_value = server_name

        self.assertEqual(DiscordUser.objects.server_name(), server_name)
        self.assertFalse(mock_logger.warning.called)

    def test_returns_empty_string_when_api_throws_http_error(
        self, mock_logger, mock_DiscordClient
    ):
        mock_exception = HTTPError('Test exception')
        mock_exception.response = Mock(**{"status_code": 440})        
        mock_DiscordClient.return_value.guild_name.side_effect = mock_exception

        self.assertEqual(DiscordUser.objects.server_name(), "")
        self.assertFalse(mock_logger.warning.called)

    def test_returns_empty_string_when_api_throws_service_error(
        self, mock_logger, mock_DiscordClient
    ):
        mock_DiscordClient.return_value.guild_name.side_effect = DiscordApiBackoff(1000)

        self.assertEqual(DiscordUser.objects.server_name(), "")
        self.assertFalse(mock_logger.warning.called)

    def test_returns_empty_string_when_api_throws_unexpected_error(
        self, mock_logger, mock_DiscordClient
    ):
        mock_DiscordClient.return_value.guild_name.side_effect = RuntimeError

        self.assertEqual(DiscordUser.objects.server_name(), "")
        self.assertTrue(mock_logger.warning.called)


@patch(MODULE_PATH + '.managers.DiscordClient', spec=DiscordClient)
class TestRoleForGroup(TestCase):    
    def test_return_role_if_found(self, mock_DiscordClient):
        mock_DiscordClient.return_value.match_role_from_name.return_value = ROLE_ALPHA

        group = Group.objects.create(name='alpha')
        self.assertEqual(DiscordUser.objects.group_to_role(group), ROLE_ALPHA)

    def test_return_empty_dict_if_not_found(self, mock_DiscordClient):
        mock_DiscordClient.return_value.match_role_from_name.return_value = dict()

        group = Group.objects.create(name='unknown')
        self.assertEqual(DiscordUser.objects.group_to_role(group), dict())



from unittest import mock

from django.contrib.auth.models import User, Group
from django.test import TestCase

from allianceauth.tests.auth_utils import AuthUtils
from allianceauth.eveonline.models import (
    EveCorporationInfo, EveAllianceInfo, EveCharacter
)

from ..models import GroupRequest, RequestLog


def create_testdata():    
    # clear DB
    User.objects.all().delete()
    Group.objects.all().delete()
    EveCharacter.objects.all().delete()
    EveCorporationInfo.objects.all().delete()
    EveAllianceInfo.objects.all().delete()

    # group 1
    group = Group.objects.create(name='Superheros')        
    group.authgroup.description = 'Default Group'
    group.authgroup.internal = False
    group.authgroup.hidden = False
    group.authgroup.save()
    
    # user 1        
    user_1 = AuthUtils.create_user('Bruce Wayne')
    AuthUtils.add_main_character_2(
        user_1,
        name='Bruce Wayne',
        character_id=1001,
        corp_id=2001,
        corp_name='Wayne Technologies'
    )        
    user_1.groups.add(group)
    group.authgroup.group_leaders.add(user_1)

    # user 2
    user_2 = AuthUtils.create_user('Clark Kent')
    AuthUtils.add_main_character_2(
        user_2,
        name='Clark Kent',
        character_id=1002,
        corp_id=2002,
        corp_name='Wayne Technologies'
    )            
    return group, user_1, user_2



class TestGroupRequest(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.group, cls.user_1, _ = create_testdata()

    def test_main_char(self):
        group_request = GroupRequest.objects.create(
            status='Pending',
            user=self.user_1,
            group=self.group
        )
        expected = self.user_1.profile.main_character
        self.assertEqual(group_request.main_char, expected)

    def test_str(self):
        group_request = GroupRequest.objects.create(
            status='Pending',
            user=self.user_1,
            group=self.group
        )
        expected = 'Bruce Wayne:Superheros'
        self.assertEqual(str(group_request), expected)


class TestRequestLog(TestCase):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.group, cls.user_1, cls.user_2 = create_testdata()

    def test_requestor(self):
        request_log = RequestLog.objects.create(
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Clark Kent'
        self.assertEqual(request_log.requestor(), expected)

    def test_type_to_str_removed(self):
        request_log = RequestLog.objects.create(
            request_type=None,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Removed'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_type_to_str_leave(self):
        request_log = RequestLog.objects.create(
            request_type=True,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Leave'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_type_to_str_join(self):
        request_log = RequestLog.objects.create(
            request_type=False,
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1
        )
        expected = 'Join'
        self.assertEqual(request_log.type_to_str(), expected)

    def test_action_to_str_accept(self):
        request_log = RequestLog.objects.create(            
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action = True
        )
        expected = 'Accept'
        self.assertEqual(request_log.action_to_str(), expected)

    def test_action_to_str_reject(self):
        request_log = RequestLog.objects.create(            
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action = False
        )
        expected = 'Reject'
        self.assertEqual(request_log.action_to_str(), expected)

    def test_req_char(self):
        request_log = RequestLog.objects.create(            
            group=self.group,
            request_info='Clark Kent:Superheros',
            request_actor=self.user_1,
            action = False
        )
        expected = self.user_2.profile.main_character
        self.assertEqual(request_log.req_char(), expected)


class TestAuthGroup(TestCase):

    def test_str(self):
        group = Group.objects.create(name='Superheros')        
        group.authgroup.description = 'Default Group'
        group.authgroup.internal = False
        group.authgroup.hidden = False
        group.authgroup.save()

        expected = 'Superheros'
        self.assertEqual(str(group.authgroup), expected)



import logging

from django.contrib.auth.models import User, Group, Permission
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models.signals import m2m_changed
from django.db.models.signals import pre_delete
from django.db.models.signals import pre_save
from django.dispatch import receiver
from .hooks import ServicesHook
from .tasks import disable_user

from allianceauth.authentication.models import State, UserProfile
from allianceauth.authentication.signals import state_changed
from allianceauth.eveonline.models import EveCharacter

logger = logging.getLogger(__name__)


@receiver(m2m_changed, sender=User.groups.through)
def m2m_changed_user_groups(sender, instance, action, *args, **kwargs):
    logger.debug("Received m2m_changed from %s groups with action %s" % (instance, action))

    def trigger_service_group_update():
        logger.debug("Triggering service group update for %s" % instance)
        # Iterate through Service hooks
        for svc in ServicesHook.get_services():
            try:
                svc.validate_user(instance)
                svc.update_groups(instance)
            except:
                logger.exception('Exception running update_groups for services module %s on user %s' % (svc, instance))

    if instance.pk and (action == "post_add" or action == "post_remove" or action == "post_clear"):
        logger.debug("Waiting for commit to trigger service group update for %s" % instance)
        transaction.on_commit(trigger_service_group_update)


@receiver(m2m_changed, sender=User.user_permissions.through)
def m2m_changed_user_permissions(sender, instance, action, *args, **kwargs):
    logger.debug("Received m2m_changed from user %s permissions with action %s" % (instance, action))
    logger.debug('sender: %s' % sender)
    if instance.pk and (action == "post_remove" or action == "post_clear"):
        logger.debug("Permissions changed for user {}, re-validating services".format(instance))
        # Checking permissions for a single user is quite fast, so we don't need to validate
        # That the permissions is a service permission, unlike groups.

        def validate_all_services():
            logger.debug("Validating all services for user {}".format(instance))
            for svc in ServicesHook.get_services():
                try:
                    svc.validate_user(instance)
                except:
                    logger.exception(
                        'Exception running validate_user for services module {} on user {}'.format(svc, instance))

        transaction.on_commit(lambda: validate_all_services())


@receiver(m2m_changed, sender=Group.permissions.through)
def m2m_changed_group_permissions(sender, instance, action, pk_set, *args, **kwargs):
    logger.debug("Received m2m_changed from group %s permissions with action %s" % (instance, action))
    if instance.pk and (action == "post_remove" or action == "post_clear"):
        logger.debug("Checking if service permission changed for group {}".format(instance))
        # As validating an entire groups service could lead to many thousands of permission checks
        # first we check that one of the permissions changed is, in fact, a service permission.
        perms = Permission.objects.filter(pk__in=pk_set)
        got_change = False
        service_perms = [svc.access_perm for svc in ServicesHook.get_services()]
        for perm in perms:
            natural_key = perm.natural_key()
            path_perm = "{}.{}".format(natural_key[1], natural_key[0])
            if path_perm not in service_perms:
                # Not a service permission, keep searching
                continue
            for svc in ServicesHook.get_services():
                if svc.access_perm == path_perm:
                    logger.debug("Permissions changed for group {} on "
                                 "service {}, re-validating services for groups users".format(instance, svc))

                    def validate_all_groups_users_for_service():
                        logger.debug("Performing validation for service {}".format(svc))
                        for user in instance.user_set.all():
                            svc.validate_user(user)

                    transaction.on_commit(validate_all_groups_users_for_service)
                    got_change = True
                    break  # Found service, break out of services iteration and go back to permission iteration
        if not got_change:
            logger.debug("Permission change for group {} was not service permission, ignoring".format(instance))


@receiver(m2m_changed, sender=State.permissions.through)
def m2m_changed_state_permissions(sender, instance, action, pk_set, *args, **kwargs):
    logger.debug("Received m2m_changed from state %s permissions with action %s" % (instance, action))
    if instance.pk and (action == "post_remove" or action == "post_clear"):
        logger.debug("Checking if service permission changed for state {}".format(instance))
        # As validating an entire groups service could lead to many thousands of permission checks
        # first we check that one of the permissions changed is, in fact, a service permission.
        perms = Permission.objects.filter(pk__in=pk_set)
        got_change = False
        service_perms = [svc.access_perm for svc in ServicesHook.get_services()]
        for perm in perms:
            natural_key = perm.natural_key()
            path_perm = "{}.{}".format(natural_key[1], natural_key[0])
            if path_perm not in service_perms:
                # Not a service permission, keep searching
                continue
            for svc in ServicesHook.get_services():
                if svc.access_perm == path_perm:
                    logger.debug("Permissions changed for state {} on "
                                 "service {}, re-validating services for state users".format(instance, svc))

                    def validate_all_state_users_for_service():
                        logger.debug("Performing validation for service {}".format(svc))
                        for profile in instance.userprofile_set.all():
                            svc.validate_user(profile.user)

                    transaction.on_commit(validate_all_state_users_for_service)
                    got_change = True
                    break  # Found service, break out of services iteration and go back to permission iteration
        if not got_change:
            logger.debug("Permission change for state {} was not service permission, ignoring".format(instance))


@receiver(state_changed)
def check_service_accounts_state_changed(sender, user, state, **kwargs):
    logger.debug("Received state_changed from %s to state %s" % (user, state))    
    for svc in ServicesHook.get_services():
        svc.validate_user(user)
        svc.update_groups(user)


@receiver(pre_delete, sender=User)
def pre_delete_user(sender, instance, *args, **kwargs):
    logger.debug("Received pre_delete from %s" % instance)
    disable_user(instance)


@receiver(pre_save, sender=User)
def disable_services_on_inactive(sender, instance, *args, **kwargs):
    logger.debug("Received pre_save from %s" % instance)
    # check if user is being marked active/inactive
    if not instance.pk:
        # new model being created
        return
    try:
        old_instance = User.objects.get(pk=instance.pk)
        if old_instance.is_active and not instance.is_active:
            logger.info("Disabling services for inactivation of user %s" % instance)
            disable_user(instance)
    except User.DoesNotExist:
        pass


@receiver(pre_save, sender=UserProfile)
def process_main_character_change(sender, instance, *args, **kwargs):    
    if not instance.pk:
        # ignore new model being created
        return
    try:
        logger.debug(
            "Received pre_save from %s for process_main_character_change", instance
        )
        old_instance = UserProfile.objects.get(pk=instance.pk)
        if old_instance.main_character and not instance.main_character:
            logger.info(
                "Disabling services due to loss of main character for user %s",
                instance.user
            )
            disable_user(instance.user)
        elif old_instance.main_character != instance.main_character:              
            logger.info(
                "Updating Names due to change of main character for user %s", 
                instance.user
            )
            for svc in ServicesHook.get_services():
                try:
                    svc.validate_user(instance.user)
                    svc.sync_nickname(instance.user)
                except:
                    logger.exception(
                        "Exception running sync_nickname for services module %s "
                        "on user %s", 
                        svc, 
                        instance
                    )

    except UserProfile.DoesNotExist:
        pass


@receiver(pre_save, sender=EveCharacter)
def process_main_character_update(sender, instance, *args, **kwargs):
    try:        
        if instance.userprofile:
            logger.debug(
                "Received pre_save from %s for process_main_character_update", 
                instance
            )
            old_instance = EveCharacter.objects.get(pk=instance.pk)
            if not instance.character_name == old_instance.character_name or \
               not instance.corporation_name == old_instance.corporation_name or \
               not instance.alliance_name == old_instance.alliance_name:
                logger.info("syncing service nickname for user {0}".format(instance.userprofile.user))

                for svc in ServicesHook.get_services():
                    try:
                        svc.validate_user(instance.userprofile.user)
                        svc.sync_nickname(instance.userprofile.user)
                    except:
                        logger.exception('Exception running sync_nickname for services module %s on user %s' % (svc, instance))

    except ObjectDoesNotExist:  # not a main char ignore
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-22 23:35
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hrapplications', '0003_applicationquestion_multi_select'),
    ]

    operations = [
        migrations.AlterField(
            model_name='hrapplication',
            name='about',
            field=models.TextField(default=''),
        ),
        migrations.AlterField(
            model_name='hrapplication',
            name='character_name',
            field=models.CharField(default='', max_length=254),
        ),
        migrations.AlterField(
            model_name='hrapplication',
            name='extra',
            field=models.TextField(default=''),
        ),
        migrations.AlterField(
            model_name='hrapplication',
            name='full_api_id',
            field=models.CharField(default='', max_length=254),
        ),
        migrations.AlterField(
            model_name='hrapplication',
            name='full_api_key',
            field=models.CharField(default='', max_length=254),
        ),
        migrations.AlterField(
            model_name='hrapplication',
            name='is_a_spi',
            field=models.CharField(default='', max_length=254),
        ),
        migrations.AlterField(
            model_name='hrapplicationcomment',
            name='comment',
            field=models.CharField(default='', max_length=254),
        ),
    ]



from django.test import TestCase

from ...models import EveCharacter, EveCorporationInfo, EveAllianceInfo
from .. import eveimageserver, evewho, dotlan, zkillboard
from ...templatetags import evelinks


class TestTemplateTags(TestCase):

    def setUp(self):
        self.my_character = EveCharacter.objects.create(
            character_id=1001,
            character_name='Bruce Wayne',
            corporation_id=2001,
            corporation_name='Dummy Corporation 1',
            corporation_ticker='DC1',
            alliance_id=3001,
            alliance_name='Dummy Alliance 1',
            alliance_ticker='DA1',
        )
        self.my_character_2 = EveCharacter.objects.create(
            character_id=1002,
            character_name='Peter Parker',
            corporation_id=2002,
            corporation_name='Dummy Corporation 2',
            corporation_ticker='DC2',
        )
        self.my_alliance = EveAllianceInfo.objects.create(
            alliance_id=3001,
            alliance_name='Dummy Alliance 1',
            alliance_ticker='DA1',
            executor_corp_id=2001
        )
        self.my_corporation = EveCorporationInfo(
            corporation_id=2001,
            corporation_name='Dummy Corporation 1',
            corporation_ticker='DC1',
            member_count=42,
            alliance=self.my_alliance
        )
        
        self.my_region_id = 8001
        self.my_region_name = 'Southpark'
        
        self.my_solar_system_id = 9001
        self.my_solar_system_name = 'Gotham'
    

    def test_evewho_character_url(self):                
        self.assertEqual(
            evelinks.evewho_character_url(self.my_character),
            evewho.character_url(self.my_character.character_id),
        )
        self.assertEqual(
            evelinks.evewho_character_url(None),
            ''
        )
        self.assertEqual(
            evelinks.evewho_character_url(self.my_character.character_id),
            evewho.character_url(self.my_character.character_id),
        )
        
    
    def test_evewho_corporation_url(self):                
        self.assertEqual(
            evelinks.evewho_corporation_url(self.my_character),
            evewho.corporation_url(self.my_character.corporation_id),
        )
        self.assertEqual(
            evelinks.evewho_corporation_url(self.my_corporation),
            evewho.corporation_url(self.my_corporation.corporation_id),
        )
        self.assertEqual(
            evelinks.evewho_corporation_url(None),
            ''
        )
        self.assertEqual(
            evelinks.evewho_corporation_url(self.my_character.corporation_id),
            evewho.corporation_url(self.my_character.corporation_id),
        )


    def test_evewho_alliance_url(self):                
        self.assertEqual(
            evelinks.evewho_alliance_url(self.my_character),
            evewho.alliance_url(self.my_character.alliance_id),
        )
        self.assertEqual(
            evelinks.evewho_alliance_url(self.my_character_2),
            '',
        )
        self.assertEqual(
            evelinks.evewho_alliance_url(self.my_alliance),
            evewho.alliance_url(self.my_alliance.alliance_id),
        )
        self.assertEqual(
            evelinks.evewho_alliance_url(None),
            ''
        )
        self.assertEqual(
            evelinks.evewho_alliance_url(self.my_character.alliance_id),
            evewho.alliance_url(self.my_character.alliance_id),
        )      

    
    # dotlan
    
    def test_dotlan_corporation_url(self):                
        self.assertEqual(
            evelinks.dotlan_corporation_url(self.my_character),
            dotlan.corporation_url(self.my_character.corporation_name),
        )
        self.assertEqual(
            evelinks.dotlan_corporation_url(self.my_corporation),
            dotlan.corporation_url(self.my_corporation.corporation_name),
        )
        self.assertEqual(
            evelinks.dotlan_corporation_url(None),
            ''
        )
        self.assertEqual(
            evelinks.dotlan_corporation_url(self.my_character.corporation_name),
            dotlan.corporation_url(self.my_character.corporation_name),
        )   


    def test_dotlan_alliance_url(self):                
        self.assertEqual(
            evelinks.dotlan_alliance_url(self.my_character),
            dotlan.alliance_url(self.my_character.alliance_name),
        )
        self.assertEqual(
            evelinks.dotlan_alliance_url(self.my_character_2),
            '',
        )
        self.assertEqual(
            evelinks.dotlan_alliance_url(self.my_alliance),
            dotlan.alliance_url(self.my_alliance.alliance_name),
        )
        self.assertEqual(
            evelinks.dotlan_alliance_url(None),
            ''
        )
        self.assertEqual(
            evelinks.dotlan_alliance_url(self.my_character.alliance_name),
            dotlan.alliance_url(self.my_character.alliance_name),
        )      

    def test_dotlan_region_url(self):                
        self.assertEqual(
            evelinks.dotlan_region_url(self.my_region_name),
            dotlan.region_url(self.my_region_name),
        )        
        self.assertEqual(
            evelinks.dotlan_region_url(None),
            ''
        )        

    def test_dotlan_solar_system_url(self):                
        self.assertEqual(
            evelinks.dotlan_solar_system_url(self.my_solar_system_name),
            dotlan.solar_system_url(self.my_solar_system_name),
        )        
        self.assertEqual(
            evelinks.dotlan_solar_system_url(None),
            ''
        )   

    
    # zkillboard

    def test_zkillboard_character_url(self):                
        self.assertEqual(
            evelinks.zkillboard_character_url(self.my_character),
            zkillboard.character_url(self.my_character.character_id),
        )
        self.assertEqual(
            evelinks.zkillboard_character_url(None),
            ''
        )
        self.assertEqual(
            evelinks.zkillboard_character_url(self.my_character.character_id),
            zkillboard.character_url(self.my_character.character_id),
        )
        
    
    def test_zkillboard_corporation_url(self):                
        self.assertEqual(
            evelinks.zkillboard_corporation_url(self.my_character),
            zkillboard.corporation_url(self.my_character.corporation_id),
        )
        self.assertEqual(
            evelinks.zkillboard_corporation_url(self.my_corporation),
            zkillboard.corporation_url(self.my_corporation.corporation_id),
        )
        self.assertEqual(
            evelinks.zkillboard_corporation_url(None),
            ''
        )
        self.assertEqual(
            evelinks.zkillboard_corporation_url(self.my_character.corporation_id),
            zkillboard.corporation_url(self.my_character.corporation_id),
        )      


    def test_zkillboard_alliance_url(self):                
        self.assertEqual(
            evelinks.zkillboard_alliance_url(self.my_character),
            zkillboard.alliance_url(self.my_character.alliance_id),
        )
        self.assertEqual(
            evelinks.zkillboard_alliance_url(self.my_character_2),
            '',
        )
        self.assertEqual(
            evelinks.zkillboard_alliance_url(self.my_alliance),
            zkillboard.alliance_url(self.my_alliance.alliance_id),
        )
        self.assertEqual(
            evelinks.zkillboard_alliance_url(None),
            ''
        )
        self.assertEqual(
            evelinks.zkillboard_alliance_url(self.my_character.alliance_id),
            zkillboard.alliance_url(self.my_character.alliance_id),
        )        


    def test_zkillboard_region_url(self):                
        self.assertEqual(
            evelinks.zkillboard_region_url(self.my_region_id),
            zkillboard.region_url(self.my_region_id),
        )        
        self.assertEqual(
            evelinks.zkillboard_region_url(None),
            ''
        )        


    def test_zkillboard_solar_system_url(self):                
        self.assertEqual(
            evelinks.zkillboard_solar_system_url(self.my_solar_system_id),
            zkillboard.solar_system_url(self.my_solar_system_id),
        )        
        self.assertEqual(
            evelinks.zkillboard_solar_system_url(None),
            ''
        )   


    # image URLs

    def test_character_portrait_url(self):
        self.assertEqual(
            evelinks.character_portrait_url(123),
            EveCharacter.generic_portrait_url(123)
            
        ),
        self.assertEqual(
            evelinks.character_portrait_url(123, 128),
            EveCharacter.generic_portrait_url(123, 128)
            
        )
        self.assertEqual(
            evelinks.character_portrait_url(123, 99),
            ''
        )
        self.assertEqual(
            evelinks.character_portrait_url(self.my_character),
            self.my_character.portrait_url()  
        )
        self.assertEqual(
            evelinks.character_portrait_url(None),
            ''
        )


    def test_corporation_logo_url(self):
        self.assertEqual(
            evelinks.corporation_logo_url(123),
            EveCorporationInfo.generic_logo_url(123)
        ),
        self.assertEqual(
            evelinks.corporation_logo_url(123, 128),
            EveCorporationInfo.generic_logo_url(123, 128)
        )
        self.assertEqual(
            evelinks.corporation_logo_url(123, 99),
            ''            
        )
        self.assertEqual(
            evelinks.corporation_logo_url(self.my_corporation),
            self.my_corporation.logo_url()
        )
        self.assertEqual(
            evelinks.corporation_logo_url(self.my_character),
            self.my_character.corporation_logo_url()
        )
        self.assertEqual(
            evelinks.corporation_logo_url(None),
            ''
        )


    def test_alliance_logo_url(self):
        self.assertEqual(            
            evelinks.alliance_logo_url(123),
            EveAllianceInfo.generic_logo_url(123)
        ),
        self.assertEqual(
            evelinks.alliance_logo_url(123, 128),
            EveAllianceInfo.generic_logo_url(123, 128)
        )
        self.assertEqual(
            evelinks.alliance_logo_url(123, 99),
            ''
        )   
        self.assertEqual(
            evelinks.alliance_logo_url(self.my_alliance),
            self.my_alliance.logo_url()
        )
        self.assertEqual(
            evelinks.alliance_logo_url(self.my_character),
            self.my_character.alliance_logo_url()
        )
        self.assertEqual(
            evelinks.alliance_logo_url(None),
            ''
        )
        self.assertEqual(
            evelinks.alliance_logo_url(self.my_character_2),
            ''
        )

    def test_type_icon_url(self):
        expected = eveimageserver.type_icon_url(123)
        self.assertEqual(evelinks.type_icon_url(123), expected)

        expected = eveimageserver.type_icon_url(123, 128)
        self.assertEqual(evelinks.type_icon_url(123, 128), expected)
        
        expected = ''
        self.assertEqual(evelinks.type_icon_url(123, 99), expected)
        
        expected = ''
        self.assertEqual(evelinks.type_icon_url(None), expected)

    def test_type_render_url(self):
        expected = eveimageserver.type_render_url(123)
        self.assertEqual(evelinks.type_render_url(123), expected)

        expected = eveimageserver.type_render_url(123, 128)
        self.assertEqual(evelinks.type_render_url(123, 128), expected)
        
        expected = ''
        self.assertEqual(evelinks.type_render_url(123, 99), expected)
        
        expected = ''
        self.assertEqual(evelinks.type_render_url(None), expected)


from django.core.checks import Error
from django.conf import settings


def check_login_scopes_setting(*args, **kwargs):
    errors = []
    try:
        assert settings.LOGIN_TOKEN_SCOPES
    except (AssertionError, AttributeError):
        errors.append(Error('LOGIN_TOKEN_SCOPES setting cannot be blank.',
                            hint='SSO tokens used for logging in must require scopes to be refreshable.'))
    return errors



from django.contrib.auth.models import Group
from allianceauth.tests.auth_utils import AuthUtils
from ..discord_client.tests import (    # noqa
    TEST_GUILD_ID,
    TEST_USER_ID,
    TEST_USER_NAME,
    TEST_USER_DISCRIMINATOR,
    create_role, 
    ROLE_ALPHA, 
    ROLE_BRAVO, 
    ROLE_CHARLIE, 
    ROLE_MIKE, 
    ALL_ROLES,
    create_user_info
)   

DEFAULT_AUTH_GROUP = 'Member'
MODULE_PATH = 'allianceauth.services.modules.discord'

TEST_MAIN_NAME = 'Spiderman'
TEST_MAIN_ID = 1005


def add_permissions_to_members():
    permission = AuthUtils.get_permission_by_name('discord.access_discord')
    members = Group.objects.get_or_create(name=DEFAULT_AUTH_GROUP)[0]
    AuthUtils.add_permissions_to_groups([permission], [members])



from django import forms
from django.contrib import admin

from allianceauth import hooks
from allianceauth.authentication.admin import (
    user_profile_pic, 
    user_username, 
    user_main_organization, 
    MainCorporationsFilter,
    MainAllianceFilter
)

from .models import NameFormatConfig


class ServicesUserAdmin(admin.ModelAdmin):
    """Parent class for UserAdmin classes for all services"""
    class Media:
        css = {
            "all": ("services/admin.css",)
        }

    search_fields = ('user__username',)
    ordering = ('user__username',)
    list_select_related = True              
    list_display = (
        user_profile_pic,
        user_username,
        '_state',
        user_main_organization, 
        '_date_joined'
    )
    list_filter = (        
        'user__profile__state',
        MainCorporationsFilter,        
        MainAllianceFilter,
        'user__date_joined',
    )

    def _state(self, obj):
        return obj.user.profile.state.name

    _state.short_description = 'state'
    _state.admin_order_field = 'user__profile__state__name'

    def _date_joined(self, obj):
        return obj.user.date_joined
    
    _date_joined.short_description = 'date joined'
    _date_joined.admin_order_field = 'user__date_joined'


class NameFormatConfigForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(NameFormatConfigForm, self).__init__(*args, **kwargs)
        SERVICE_CHOICES = \
            [(s.name, s.name) for h in hooks.get_hooks('services_hook') for s in [h()]]
        if self.instance.id:
            current_choice = (self.instance.service_name, self.instance.service_name)
            if current_choice not in SERVICE_CHOICES:
                SERVICE_CHOICES.append(current_choice)
        self.fields['service_name'] = forms.ChoiceField(choices=SERVICE_CHOICES)


class NameFormatConfigAdmin(admin.ModelAdmin):
    form = NameFormatConfigForm
    list_display = ('service_name', 'get_state_display_string')

    def get_state_display_string(self, obj):
        return ', '.join([state.name for state in obj.states.all()])
    get_state_display_string.short_description = 'States'


admin.site.register(NameFormatConfig, NameFormatConfigAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-22 23:35
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fleetactivitytracking', '0003_auto_20160906_2354'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fatlink',
            name='fleet',
            field=models.CharField(default='', max_length=254),
        ),
    ]



"""
Abstract view classes for building services.

These view classes are provided as convenience only. If they
don't make sense to use in your service, there is no obligation
to use these views. You are free to build the internal structure
of the service as you like.
"""

from collections import OrderedDict
from django.views import View
from django.urls import reverse_lazy
from django.views.generic import UpdateView, DeleteView
from django.views.generic.detail import SingleObjectMixin
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.db import models, IntegrityError
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import render, Http404, redirect
from django.utils.translation import gettext_lazy as _

from .forms import ServicePasswordModelForm

import logging

logger = logging.getLogger(__name__)


class AbstractServiceModel(models.Model):
    user = models.OneToOneField('auth.User',
                                primary_key=True,
                                on_delete=models.CASCADE,
                                related_name='%(app_label)s'
                                )

    def __init__(self, *args, **kwargs):
        super(AbstractServiceModel, self).__init__(*args, **kwargs)
        self.credentials = OrderedDict()
        # Should be set with a dict of service credentials (username, password etc) when changed

    def update_password(self, password=None):
        pass

    def reset_password(self):
        pass

    class Meta:
        abstract = True


class BaseServiceView(LoginRequiredMixin, PermissionRequiredMixin, View):
    """
    Define:
        permission_required
    """
    index_redirect = 'services:services'
    success_url = reverse_lazy(index_redirect)
    model = AbstractServiceModel  # Overload
    service_name = 'base'  # Overload


class ServiceCredentialsViewMixin:
    template_name = 'services/service_credentials.html'


class BaseCreatePasswordServiceAccountView(BaseServiceView, ServiceCredentialsViewMixin):
    def get(self, request):
        logger.debug("{} called by user {}".format(self.__class__.__name__, request.user))
        try:
            svc_obj = self.model.objects.create(user=request.user)
        except IntegrityError:
            messages.error(request, _("That service account already exists"))
            return redirect(self.index_redirect)

        return render(request, self.template_name,
                      context={'credentials': svc_obj.credentials, 'service': self.service_name, 'view': self})


class ServicesCRUDMixin(SingleObjectMixin):
    def get_object(self, queryset=None):
        """
        Returns the object the view is displaying.
        """
        if queryset is None:
            queryset = self.get_queryset()

        try:
            return queryset.get(user__pk=self.request.user.pk)
        except ObjectDoesNotExist:
            raise Http404


class BaseDeactivateServiceAccountView(ServicesCRUDMixin, BaseServiceView, DeleteView):
    template_name = 'services/service_confirm_delete.html'


class BaseSetPasswordServiceAccountView(ServicesCRUDMixin, BaseServiceView, UpdateView):
    template_name = 'services/service_password.html'
    form_class = ServicePasswordModelForm  # You should overload this with a subclass

    def post(self, request, *args, **kwargs):
        result = super(BaseSetPasswordServiceAccountView, self).post(request, *args, **kwargs)
        if self.get_form().is_valid():
            messages.success(request, _("Successfully set your {} password".format(self.service_name)))
        return result


class BaseResetPasswordServiceAccountView(ServicesCRUDMixin, BaseServiceView, ServiceCredentialsViewMixin):
    """
    Set a random password
    """
    def get(self, request):
        svc_obj = self.get_object()
        svc_obj.reset_password()
        return render(request, self.template_name,
                      context={'credentials': svc_obj.credentials, 'service': self.service_name, 'view': self})



from allianceauth.services.hooks import MenuItemHook, UrlHook
from django.utils.translation import ugettext_lazy as _
from allianceauth import hooks
from allianceauth.corputils import urls


class CorpStats(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(self,
                              _('Corporation Stats'),
                              'fas fa-share-alt fa-fw',
                              'corputils:view',
                              navactive=['corputils:'])

    def render(self, request):
        if request.user.has_perm('corputils.view_corp_corpstats') or request.user.has_perm(
                'corputils.view_alliance_corpstats') or request.user.has_perm(
                'corputils.add_corpstats') or request.user.has_perm('corputils.view_state_corpstats'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return CorpStats()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'corputils', r'^corpstats/')



from django.conf import settings


def _clean_setting(
    name: str,     
    default_value: object,         
    min_value: int = None,
    max_value: int = None,
    required_type: type = None
):
    """cleans the input for a custom setting
    
    Will use `default_value` if settings does not exit or has the wrong type
    or is outside define boundaries (for int only)

    Need to define `required_type` if `default_value` is `None`

    Will assume `min_value` of 0 for int (can be overriden)

    Returns cleaned value for setting
    """                
    if default_value is None and not required_type:
        raise ValueError('You must specify a required_type for None defaults')
    
    if not required_type:
        required_type = type(default_value)

    if min_value is None and required_type == int:
        min_value = 0        
        
    if (hasattr(settings, name)
        and isinstance(getattr(settings, name), required_type)
        and (min_value is None or getattr(settings, name) >= min_value)
        and (max_value is None or getattr(settings, name) <= max_value)
    ):        
        return getattr(settings, name)    
    else:
        return default_value


AUTHENTICATION_ADMIN_USERS_MAX_GROUPS = \
    _clean_setting('AUTHENTICATION_ADMIN_USERS_MAX_GROUPS', 10)

AUTHENTICATION_ADMIN_USERS_MAX_CHARS = \
    _clean_setting('AUTHENTICATION_ADMIN_USERS_MAX_CHARS', 5)




# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-07 19:14
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authservicesinfo',
            name='pathfinder_password',
        ),
        migrations.RemoveField(
            model_name='authservicesinfo',
            name='pathfinder_username',
        ),
    ]



import logging

import requests
import amqp.exceptions
from packaging.version import Version as Pep440Version, InvalidVersion
from celery.app import app_or_default

from django import template
from django.conf import settings
from django.core.cache import cache

from allianceauth import __version__


register = template.Library()

# cache timers
TAG_CACHE_TIME = 3600  # 1 hours
NOTIFICATION_CACHE_TIME = 300  # 5 minutes
# timeout for all requests
REQUESTS_TIMEOUT = 5    # 5 seconds
# max pages to be fetched from gitlab
MAX_PAGES = 50

GITLAB_AUTH_REPOSITORY_TAGS_URL = (
    'https://gitlab.com/api/v4/projects/allianceauth%2Fallianceauth/repository/tags'
)
GITLAB_AUTH_ANNOUNCEMENT_ISSUES_URL = (
    'https://gitlab.com/api/v4/projects/allianceauth%2Fallianceauth/issues'
    '?labels=announcement'
)

logger = logging.getLogger(__name__)


@register.inclusion_tag('allianceauth/admin-status/overview.html')
def status_overview() -> dict:
    response = {
        'notifications': list(),        
        'current_version': __version__,
        'task_queue_length': -1,
    }
    response.update(_current_notifications())
    response.update(_current_version_summary())
    response.update({'task_queue_length': _fetch_celery_queue_length()})
    return response


def _fetch_celery_queue_length() -> int:
    try:
        app = app_or_default(None)
        with app.connection_or_acquire() as conn:
            return conn.default_channel.queue_declare(
                queue=getattr(settings, 'CELERY_DEFAULT_QUEUE', 'celery'), 
                passive=True
            ).message_count
    except amqp.exceptions.ChannelError:
        # Queue doesn't exist, probably empty
        return 0
    except Exception:
        logger.exception("Failed to get celery queue length")
    return -1


def _current_notifications() -> dict:
    """returns the newest 5 announcement issues"""    
    try:
        notifications = cache.get_or_set(
            'gitlab_notification_issues', 
            _fetch_notification_issues_from_gitlab,
            NOTIFICATION_CACHE_TIME
        )                
    except requests.RequestException:
        logger.exception('Error while getting gitlab notifications')
        top_notifications = []
    else:
        if notifications:
            top_notifications = notifications[:5]
        else:
            top_notifications = []
    
    response = {
        'notifications': top_notifications,
    }
    return response


def _fetch_notification_issues_from_gitlab() -> list:    
    return _fetch_list_from_gitlab(GITLAB_AUTH_ANNOUNCEMENT_ISSUES_URL, max_pages=10)


def _current_version_summary() -> dict:
    """returns the current version info"""    
    try:
        tags = cache.get_or_set(
            'git_release_tags', _fetch_tags_from_gitlab, TAG_CACHE_TIME
        )
    except requests.RequestException:
        logger.exception('Error while getting gitlab release tags')
        return {}
        
    if not tags:
        return {}
        
    (
        latest_major_version, 
        latest_minor_version, 
        latest_patch_version, 
        latest_beta_version
    ) = _latests_versions(tags)    
    current_version = Pep440Version(__version__)

    has_latest_major = \
        current_version >= latest_major_version if latest_major_version else False
    has_latest_minor = \
        current_version >= latest_minor_version if latest_minor_version else False
    has_latest_patch = \
        current_version >= latest_patch_version if latest_patch_version else False
    has_current_beta = \
        current_version.base_version <= latest_beta_version.base_version \
        and latest_major_version.base_version <= latest_beta_version.base_version \
        if latest_beta_version else False

    response = {        
        'latest_major': has_latest_major,
        'latest_minor': has_latest_minor,
        'latest_patch': has_latest_patch,
        'latest_beta': has_current_beta,
        'current_version': str(current_version),
        'latest_major_version': str(latest_major_version),
        'latest_minor_version': str(latest_minor_version),
        'latest_patch_version': str(latest_patch_version),
        'latest_beta_version': str(latest_beta_version)
    }        
    return response


def _fetch_tags_from_gitlab():
    return _fetch_list_from_gitlab(GITLAB_AUTH_REPOSITORY_TAGS_URL)


def _latests_versions(tags: list) -> tuple:
    """returns latests version from given tags list
    
    Non-compliant tags will be ignored
    """
    versions = list()
    betas = list()
    for tag in tags:
        try:
            version = Pep440Version(tag.get('name'))            
        except InvalidVersion:
            pass
        else:
            if version.is_prerelease or version.is_devrelease:
                betas.append(version)
            else:
                versions.append(version)

    latest_version = latest_patch_version = max(versions)
    latest_major_version = min([
        v for v in versions if v.major == latest_version.major
    ])
    latest_minor_version = min([
        v for v in versions 
        if v.major == latest_version.major and v.minor == latest_version.minor
    ])
    latest_beta_version = max(betas)
    return (
        latest_major_version, 
        latest_minor_version, 
        latest_patch_version, 
        latest_beta_version
    )


def _fetch_list_from_gitlab(url: str, max_pages: int = MAX_PAGES) -> list:
    """returns a list from the GitLab API. Supports pageing"""
    result = list()    
    for page in range(1, max_pages + 1):
        request = requests.get(
            url, params={'page': page}, timeout=REQUESTS_TIMEOUT
        )
        request.raise_for_status()
        result += request.json()
        if 'x-total-pages' in request.headers:
            try:
                total_pages = int(request.headers['x-total-pages'])
            except ValueError:
                total_pages = None
        else:
            total_pages = None
            
        if not total_pages or page >= total_pages:
            break
    
    return result



from django.contrib import admin

from .models import DiscourseUser
from ...admin import ServicesUserAdmin


@admin.register(DiscourseUser)
class DiscourseUserAdmin(ServicesUserAdmin):
    list_display = ServicesUserAdmin.list_display + (        
        'enabled',        
    ) 



from django.utils.translation import ugettext_lazy as _

from allianceauth.services.hooks import MenuItemHook, UrlHook
from allianceauth import hooks

from . import urls
from .managers import GroupManager


class GroupManagementMenuItem(MenuItemHook):
    """ This class ensures only authorized users will see the menu entry """
    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=_('Group Management'),
            classes='fas fa-users-cog fa-fw',
            url_name='groupmanagement:management',
            order=50,
            navactive=['groupmanagement:management']
        )

    def render(self, request):
        if GroupManager.can_manage_groups(request.user):
            app_count = GroupManager.pending_requests_count_for_user(request.user)
            self.count = app_count if app_count and app_count > 0 else None
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return GroupManagementMenuItem()


@hooks.register('url_hook')
def register_urls():
    return UrlHook(urls, 'group', r'^group/')



from django import forms
from django.utils.translation import ugettext_lazy as _
from allianceauth.authentication.models import User
class RegistrationForm(forms.Form):
    email = forms.EmailField(label=_('Email'), max_length=254, required=True)

    class _meta:
        model = User



from django.conf import settings

from django.contrib import admin
from django.contrib.auth.models import Group as BaseGroup, User
from django.db.models import Count
from django.db.models.functions import Lower
from django.db.models.signals import pre_save, post_save, pre_delete, \
    post_delete, m2m_changed
from django.dispatch import receiver

from .models import AuthGroup
from .models import GroupRequest
from . import signals

if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True    
else:
    _has_auto_groups = False


class AuthGroupInlineAdmin(admin.StackedInline):
    model = AuthGroup
    filter_horizontal = ('group_leaders', 'group_leader_groups', 'states',)
    fields = (
        'description', 
        'group_leaders', 
        'group_leader_groups', 
        'states', 'internal', 
        'hidden', 
        'open', 
        'public'
    )
    verbose_name_plural = 'Auth Settings'
    verbose_name = ''

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """overriding this formfield to have sorted lists in the form"""
        if db_field.name == "group_leaders":
            kwargs["queryset"] = User.objects.order_by(Lower('username'))
        elif db_field.name == "group_leader_groups":
            kwargs["queryset"] = Group.objects.order_by(Lower('name'))
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return request.user.has_perm('auth.change_group')


if _has_auto_groups:
    class IsAutoGroupFilter(admin.SimpleListFilter):
        title = 'auto group'
        parameter_name = 'is_auto_group__exact'

        def lookups(self, request, model_admin):
            return (
                ('yes', 'Yes'),
                ('no', 'No'),
            )

        def queryset(self, request, queryset):
            value = self.value()
            if value == 'yes':
                return queryset.exclude(
                    managedalliancegroup__isnull=True, 
                    managedcorpgroup__isnull=True
                )
            elif value == 'no':
                return queryset.filter(
                    managedalliancegroup__isnull=True, 
                    managedcorpgroup__isnull=True
                )
            else:
                return queryset


class HasLeaderFilter(admin.SimpleListFilter):
    title = 'has leader'
    parameter_name = 'has_leader__exact'

    def lookups(self, request, model_admin):
        return (
            ('yes', 'Yes'),
            ('no', 'No'),
        )

    def queryset(self, request, queryset):
        value = self.value()
        if value == 'yes':
            return queryset.filter(authgroup__group_leaders__isnull=False)
        elif value == 'no':
            return queryset.filter(authgroup__group_leaders__isnull=True)
        else:
            return queryset

class GroupAdmin(admin.ModelAdmin):    
    list_select_related = True
    ordering = ('name', )
    list_display = (
        'name', 
        '_description', 
        '_properties', 
        '_member_count', 
        'has_leader'
    )
    list_filter = [
        'authgroup__internal', 
        'authgroup__hidden', 
        'authgroup__open', 
        'authgroup__public',        
    ]
    if _has_auto_groups:
        list_filter.append(IsAutoGroupFilter)
    list_filter.append(HasLeaderFilter)

    search_fields = ('name', 'authgroup__description')
    
    def get_queryset(self, request):
        qs = super().get_queryset(request)                
        qs = qs.annotate(
            member_count=Count('user', distinct=True),
        )        
        return qs

    def _description(self, obj):
        return obj.authgroup.description

    def _member_count(self, obj):
        return obj.member_count

    _member_count.short_description = 'Members'
    _member_count.admin_order_field = 'member_count'
    
    def has_leader(self, obj):
        return obj.authgroup.group_leaders.exists()
    
    has_leader.boolean = True    

    def _properties(self, obj):
        properties = list()       
        if _has_auto_groups and (
            obj.managedalliancegroup_set.exists() 
            or obj.managedcorpgroup_set.exists()
        ):
            properties.append('Auto Group')
        elif obj.authgroup.internal:
            properties.append('Internal')
        else:
            if obj.authgroup.hidden:
                properties.append('Hidden')
            if obj.authgroup.open:
                properties.append('Open')
            if obj.authgroup.public:
                properties.append('Public')
        if not properties:
            properties.append('Default')
                    
        return properties

    _properties.short_description = "properties"

    filter_horizontal = ('permissions',)
    inlines = (AuthGroupInlineAdmin,)


class Group(BaseGroup):
    class Meta:
        proxy = True
        verbose_name = BaseGroup._meta.verbose_name
        verbose_name_plural = BaseGroup._meta.verbose_name_plural

try:
    admin.site.unregister(BaseGroup)
finally:
    admin.site.register(Group, GroupAdmin)


admin.site.register(GroupRequest)


@receiver(pre_save, sender=Group)
def redirect_pre_save(sender, signal=None, *args, **kwargs):
    pre_save.send(BaseGroup, *args, **kwargs)


@receiver(post_save, sender=Group)
def redirect_post_save(sender, signal=None, *args, **kwargs):
    post_save.send(BaseGroup, *args, **kwargs)


@receiver(pre_delete, sender=Group)
def redirect_pre_delete(sender, signal=None, *args, **kwargs):
    pre_delete.send(BaseGroup, *args, **kwargs)


@receiver(post_delete, sender=Group)
def redirect_post_delete(sender, signal=None, *args, **kwargs):
    post_delete.send(BaseGroup, *args, **kwargs)


@receiver(m2m_changed, sender=Group.permissions.through)
def redirect_m2m_changed_permissions(sender, signal=None, *args, **kwargs):
    m2m_changed.send(BaseGroup, *args, **kwargs)



from unittest import mock

from django.test import TestCase

from ..models import EveCharacter, EveCorporationInfo, EveAllianceInfo
from ..providers import Character, Corporation, Alliance


class EveCharacterProviderManagerTestCase(TestCase):
    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_get_character(self, provider):
        expected = Character()
        provider.get_character.return_value = expected

        result = EveCharacter.provider.get_character(1234)

        self.assertEqual(expected, result)


class EveCharacterManagerTestCase(TestCase):

    class TestCharacter(Character):
        @property
        def alliance(self):
            return Alliance(id=3456, name='Test Alliance')

        @property
        def corp(self):
            return Corporation(
                id=2345, 
                name='Test Corp', 
                alliance_id=3456, 
                ticker='0BUGS' #lies, blatant lies!
            )

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_create_character(self, provider):
        # Also covers create_character_obj
        expected = self.TestCharacter(
            id=1234, 
            name='Test Character', 
            corp_id=2345, 
            alliance_id=3456
        )

        provider.get_character.return_value = expected

        result = EveCharacter.objects.create_character(1234)

        self.assertEqual(result.character_id, expected.id)
        self.assertEqual(result.character_name, expected.name)
        self.assertEqual(result.corporation_id, expected.corp.id)
        self.assertEqual(result.corporation_name, expected.corp.name)
        self.assertEqual(result.corporation_ticker, expected.corp.ticker)
        self.assertEqual(result.alliance_id, expected.alliance.id)
        self.assertEqual(result.alliance_name, expected.alliance.name)

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_update_character(self, provider):
        # Also covers Model.update_character
        existing = EveCharacter.objects.create(
            character_id=1234,
            character_name='character.name',
            corporation_id=23457,
            corporation_name='character.corp.name',
            corporation_ticker='cc1',
            alliance_id=34567,
            alliance_name='character.alliance.name',
        )

        expected = self.TestCharacter(
            id=1234, 
            name='Test Character', 
            corp_id=2345, 
            alliance_id=3456
        )
        provider.get_character.return_value = expected

        result = EveCharacter.objects.update_character(1234)

        self.assertEqual(result.character_id, expected.id)
        self.assertEqual(result.character_name, expected.name)
        self.assertEqual(result.corporation_id, expected.corp.id)
        self.assertEqual(result.corporation_name, expected.corp.name)
        self.assertEqual(result.corporation_ticker, expected.corp.ticker)
        self.assertEqual(result.alliance_id, expected.alliance.id)
        self.assertEqual(result.alliance_name, expected.alliance.name)

    def test_get_character_by_id(self):
        EveCharacter.objects.all().delete()
        EveCharacter.objects.create(
            character_id=1234,
            character_name='character.name',
            corporation_id=2345,
            corporation_name='character.corp.name',
            corporation_ticker='cc1',
            alliance_id=3456,
            alliance_name='character.alliance.name',
        )

        # try to get existing character
        result = EveCharacter.objects.get_character_by_id(1234)

        self.assertEqual(result.character_id, 1234)
        self.assertEqual(result.character_name, 'character.name')

        # try to get non existing character        
        self.assertIsNone(EveCharacter.objects.get_character_by_id(9999))


class EveAllianceProviderManagerTestCase(TestCase):
    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_get_alliance(self, provider):
        expected = Alliance()
        provider.get_alliance.return_value = expected

        result = EveAllianceInfo.provider.get_alliance(1234)

        self.assertEqual(expected, result)


class EveAllianceManagerTestCase(TestCase):

    class TestAlliance(Alliance):
        def corp(self, id):
            return self._corps[id]

    @mock.patch('allianceauth.eveonline.models.EveAllianceInfo.populate_alliance')
    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_create_alliance(self, provider, populate_alliance):
        # Also covers create_alliance_obj
        expected = self.TestAlliance(
            id=3456, 
            name='Test Alliance', 
            ticker='TEST',
            corp_ids=[2345], 
            executor_corp_id=2345
        )

        provider.get_alliance.return_value = expected

        result = EveAllianceInfo.objects.create_alliance(3456)

        self.assertEqual(result.alliance_id, expected.id)
        self.assertEqual(result.alliance_name, expected.name)
        self.assertEqual(result.alliance_ticker, expected.ticker)
        self.assertEqual(result.executor_corp_id, expected.executor_corp_id)
        self.assertTrue(populate_alliance.called)

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_update_alliance(self, provider):
        # Also covers Model.update_alliance
        EveAllianceInfo.objects.create(
            alliance_id=3456,
            alliance_name='alliance.name',
            alliance_ticker='at1',
            executor_corp_id=2345,
        )
        expected = self.TestAlliance(
            id=3456, 
            name='Test Alliance', 
            ticker='TEST',
            corp_ids=[2345], 
            executor_corp_id=2345
        )

        provider.get_alliance.return_value = expected

        result = EveAllianceInfo.objects.update_alliance(3456)

        # This is the only thing ever updated in code
        self.assertEqual(result.executor_corp_id, expected.executor_corp_id)


class EveCorporationProviderManagerTestCase(TestCase):
    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_get_corporation(self, provider):
        expected = Corporation()
        provider.get_corp.return_value = expected

        result = EveCorporationInfo.provider.get_corporation(2345)

        self.assertEqual(expected, result)


class EveCorporationManagerTestCase(TestCase):

    class TestCorporation(Corporation):
        @property
        def alliance(self):
            return EveAllianceManagerTestCase.TestAlliance(
                id=3456, 
                name='Test Alliance', 
                ticker='TEST',
                corp_ids=[2345], 
                executor_corp_id=2345
            )

        @property
        def ceo(self):
            return EveCharacterManagerTestCase.TestCharacter(
                id=1234, 
                name='Test Character',
                corp_id=2345, 
                alliance_id=3456
            )

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_create_corporation(self, provider):
        # Also covers create_corp_obj
        exp_alliance = EveAllianceInfo.objects.create(
            alliance_id=3456,
            alliance_name='alliance.name',
            alliance_ticker='99bug',
            executor_corp_id=2345,
        )

        expected = self.TestCorporation(
            id=2345, 
            name='Test Corp', 
            ticker='0BUGS',
            ceo_id=1234, 
            members=1, 
            alliance_id=3456
        )

        provider.get_corp.return_value = expected

        result = EveCorporationInfo.objects.create_corporation('2345')

        self.assertEqual(result.corporation_id, expected.id)
        self.assertEqual(result.corporation_name, expected.name)
        self.assertEqual(result.corporation_ticker, expected.ticker)
        self.assertEqual(result.member_count, expected.members)
        self.assertEqual(result.alliance, exp_alliance)

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_create_corporation_no_alliance(self, provider):
        # variant to test no alliance case
        # Also covers create_corp_obj        
        expected = self.TestCorporation(
            id=2345, 
            name='Test Corp', 
            ticker='0BUGS',
            ceo_id=1234, 
            members=1, 
            alliance_id=3456
        )

        provider.get_corp.return_value = expected

        result = EveCorporationInfo.objects.create_corporation(2345)

        self.assertEqual(result.corporation_id, expected.id)
        self.assertEqual(result.corporation_name, expected.name)
        self.assertEqual(result.corporation_ticker, expected.ticker)
        self.assertEqual(result.member_count, expected.members)
        self.assertIsNone(result.alliance)

    @mock.patch('allianceauth.eveonline.managers.providers.provider')
    def test_update_corporation(self, provider):
        # Also covers Model.update_corporation
        exp_alliance = EveAllianceInfo.objects.create(
            alliance_id=3456,
            alliance_name='alliance.name',
            alliance_ticker='at1',
            executor_corp_id=2345,
        )

        EveCorporationInfo.objects.create(
            corporation_id=2345,
            corporation_name='corp.name',
            corporation_ticker='cc1',
            member_count=10,
            alliance=None,
        )

        expected = self.TestCorporation(
            id=2345, 
            name='Test Corp', 
            ticker='0BUGS',
            ceo_id=1234, 
            members=1, 
            alliance_id=3456
        )

        provider.get_corp.return_value = expected

        result = EveCorporationInfo.objects.update_corporation('2345')

        self.assertEqual(result.corporation_id, expected.id)
        # These are the only updated props
        self.assertEqual(result.member_count, expected.members)
        self.assertEqual(result.alliance, exp_alliance)



TEST_GUILD_ID = 123456789012345678
TEST_USER_ID = 198765432012345678
TEST_USER_NAME = 'Peter Parker'
TEST_USER_DISCRIMINATOR = '1234'
TEST_BOT_TOKEN = 'abcdefhijlkmnopqastzvwxyz1234567890ABCDEFGHOJKLMNOPQRSTUVWXY'
TEST_ROLE_ID = 654321012345678912


def create_role(id: int, name: str, managed=False):
    return {
        'id': int(id),
        'name': str(name),
        'managed': bool(managed)
    }


def create_matched_role(role, created=False) -> tuple:
    return role, created


ROLE_ALPHA = create_role(1, 'alpha')
ROLE_BRAVO = create_role(2, 'bravo')
ROLE_CHARLIE = create_role(3, 'charlie')
ROLE_MIKE = create_role(13, 'mike', True)

ALL_ROLES = [ROLE_ALPHA, ROLE_BRAVO, ROLE_CHARLIE, ROLE_MIKE]


def create_user_info(
    id: int = TEST_USER_ID, 
    username: str = TEST_USER_NAME, 
    discriminator: str = TEST_USER_DISCRIMINATOR
):
    return {
        'id': str(id),
        'username': str(username[:32]),
        'discriminator': str(discriminator[:4])
    }



from unittest import mock
from django.test import TestCase
from allianceauth.notifications.context_processors import user_notification_count
from allianceauth.tests.auth_utils import AuthUtils
from django.core.cache import cache
from allianceauth.notifications.models import Notification

class TestNotificationCount(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('magic_mike')
        AuthUtils.add_main_character(cls.user, 'Magic Mike', '1', corp_id='2', corp_name='Pole Riders', corp_ticker='PRIDE', alliance_id='3', alliance_name='RIDERS')
        cls.user.profile.refresh_from_db()

        ### test notifications for mike
        Notification.objects.all().delete()
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 1 Failed",
                                    message="Because it was broken",
                                    viewed=True)
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 2 Failed",
                                    message="Because it was broken")
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 3 Failed",
                                    message="Because it was broken")
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 4 Failed",
                                    message="Because it was broken")
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 5 Failed",
                                    message="Because it was broken")
        Notification.objects.create(user=cls.user,
                                    level="INFO",
                                    title="Job 6 Failed",
                                    message="Because it was broken")

        cls.user2 = AuthUtils.create_user('teh_kid')
        AuthUtils.add_main_character(cls.user, 'The Kid', '2', corp_id='2', corp_name='Pole Riders', corp_ticker='PRIDE', alliance_id='3', alliance_name='RIDERS')
        cls.user2.profile.refresh_from_db()

        # Noitification for kid
        Notification.objects.create(user=cls.user2,
                                    level="INFO",
                                    title="Job 6 Failed",
                                    message="Because it was broken")


    def test_no_cache(self):
        mock_req = mock.MagicMock()
        mock_req.user.id = self.user.id

        cache.delete("u-note:{}".format(self.user.id)) # force the db to be hit
        context_dict = user_notification_count(mock_req)
        self.assertIsInstance(context_dict, dict)
        self.assertEqual(context_dict.get('notifications'), 5)  # 5 only


    @mock.patch('allianceauth.notifications.models.Notification.objects')
    def test_cache(self, mock_foo):
        mock_foo.filter.return_value = mock_foo
        mock_foo.count.return_value = 5
        mock_req = mock.MagicMock()
        mock_req.user.id = self.user.id

        cache.set("u-note:{}".format(self.user.id),10,5)
        context_dict = user_notification_count(mock_req)
        self.assertIsInstance(context_dict, dict)
        self.assertEqual(context_dict.get('notifications'), 10) # cached value
        self.assertEqual(mock_foo.called, 0)  # ensure the DB was not hit



# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.

__version__ = '2.8.0'
__title__ = 'Alliance Auth'
__url__ = 'https://gitlab.com/allianceauth/allianceauth'
NAME = '%s v%s' % (__title__, __version__)
default_app_config = 'allianceauth.apps.AllianceAuthConfig'



from django.test import TestCase, RequestFactory
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.utils.timezone import now

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import (
    EveCharacter, EveCorporationInfo, EveAllianceInfo
)

from ....admin import (
    user_profile_pic, 
    user_username, 
    user_main_organization, 
    ServicesUserAdmin,
    MainCorporationsFilter,
    MainAllianceFilter
)
from ..admin import DiscordUserAdmin
from ..models import DiscordUser


class TestDataMixin(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        EveCharacter.objects.all().delete()
        EveCorporationInfo.objects.all().delete()
        EveAllianceInfo.objects.all().delete()
        User.objects.all().delete()
        DiscordUser.objects.all().delete()
        
        # user 1 - corp and alliance, normal user
        cls.character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name='Bruce Wayne',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        cls.character_1a = EveCharacter.objects.create(
            character_id=1002,
            character_name='Batman',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        alliance = EveAllianceInfo.objects.create(
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',            
            executor_corp_id=2001
        )
        EveCorporationInfo.objects.create(
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',            
            member_count=42,
            alliance=alliance
        )        
        cls.user_1 = User.objects.create_user(
            cls.character_1.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_1,
            owner_hash='x1' + cls.character_1.character_name,
            user=cls.user_1
        )
        CharacterOwnership.objects.create(
            character=cls.character_1a,
            owner_hash='x1' + cls.character_1a.character_name,
            user=cls.user_1
        )
        cls.user_1.profile.main_character = cls.character_1
        cls.user_1.profile.save()
        DiscordUser.objects.create(
            user=cls.user_1,
            uid=1001,
            username='Bruce Wayne',
            discriminator='1234',
            activated=now()
        )

        # user 2 - corp only, staff
        cls.character_2 = EveCharacter.objects.create(
            character_id=1003,
            character_name='Clark Kent',
            corporation_id=2002,
            corporation_name='Daily Planet',
            corporation_ticker='DP',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2002,
            corporation_name='Daily Plane',
            corporation_ticker='DP',            
            member_count=99,
            alliance=None
        )
        cls.user_2 = User.objects.create_user(
            cls.character_2.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_2,
            owner_hash='x1' + cls.character_2.character_name,
            user=cls.user_2
        )
        cls.user_2.profile.main_character = cls.character_2
        cls.user_2.profile.save()
        DiscordUser.objects.create(
            user=cls.user_2,
            uid=1002
        )
        
        # user 3 - no main, no group, superuser
        cls.character_3 = EveCharacter.objects.create(
            character_id=1101,
            character_name='Lex Luthor',
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',            
            member_count=666,
            alliance=None
        )
        EveAllianceInfo.objects.create(
            alliance_id=3101,
            alliance_name='Lex World Domination',
            alliance_ticker='LWD',
            executor_corp_id=2101
        )
        cls.user_3 = User.objects.create_user(
            cls.character_3.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_3,
            owner_hash='x1' + cls.character_3.character_name,
            user=cls.user_3
        )
        DiscordUser.objects.create(
            user=cls.user_3,
            uid=1003
        )
        
    def setUp(self):
        self.factory = RequestFactory()
        self.modeladmin = DiscordUserAdmin(
            model=DiscordUser, admin_site=AdminSite()
        )    
    

class TestColumnRendering(TestDataMixin, TestCase):

    def test_user_profile_pic_u1(self):
        expected = (
            '<img src="https://images.evetech.net/characters/1001/'
            'portrait?size=32" class="img-circle">'
        )
        self.assertEqual(user_profile_pic(self.user_1.discord), expected)

    def test_user_profile_pic_u3(self):
        self.assertIsNone(user_profile_pic(self.user_3.discord))

    def test_user_username_u1(self):
        expected = (
            '<strong><a href="/admin/discord/discorduser/{}/change/">'
            'Bruce_Wayne</a></strong><br>Bruce Wayne'.format(
                self.user_1.discord.pk
            )
        )
        self.assertEqual(user_username(self.user_1.discord), expected)

    def test_user_username_u3(self):
        expected = (
            '<strong><a href="/admin/discord/discorduser/{}/change/">'
            'Lex_Luthor</a></strong>'.format(self.user_3.discord.pk)
        )
        self.assertEqual(user_username(self.user_3.discord), expected)

    def test_user_main_organization_u1(self):
        expected = 'Wayne Technologies<br>Wayne Enterprises'
        result = user_main_organization(self.user_1.discord)
        self.assertEqual(result, expected)

    def test_user_main_organization_u2(self):
        expected = 'Daily Planet'
        result = user_main_organization(self.user_2.discord)
        self.assertEqual(result, expected)

    def test_user_main_organization_u3(self):
        expected = None
        result = user_main_organization(self.user_3.discord)
        self.assertEqual(result, expected)

    def test_uid(self):
        expected = 1001
        result = self.modeladmin._uid(self.user_1.discord)
        self.assertEqual(result, expected)

    def test_username_when_defined(self):
        expected = 'Bruce Wayne#1234'
        result = self.modeladmin._username(self.user_1.discord)
        self.assertEqual(result, expected)

    def test_username_when_not_defined(self):
        expected = ''
        result = self.modeladmin._username(self.user_2.discord)
        self.assertEqual(result, expected)

    # actions


class TestFilters(TestDataMixin, TestCase):
    
    def test_filter_main_corporations(self):
        
        class DiscordUserAdminTest(ServicesUserAdmin): 
            list_filter = (MainCorporationsFilter,)
                
        my_modeladmin = DiscordUserAdminTest(DiscordUser, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [            
            (2002, 'Daily Planet'),
            (2001, 'Wayne Technologies'),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get(
            '/', {'main_corporation_id__exact': self.character_1.corporation_id}
        )
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [self.user_1.discord]
        self.assertSetEqual(set(queryset), set(expected))
    
    def test_filter_main_alliances(self):
        
        class DiscordUserAdminTest(ServicesUserAdmin): 
            list_filter = (MainAllianceFilter,)
                
        my_modeladmin = DiscordUserAdminTest(DiscordUser, AdminSite())
       
        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [
            (3001, 'Wayne Enterprises'),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get(
            '/', {'main_alliance_id__exact': self.character_1.alliance_id}
        )
        request.user = self.user_1                
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [self.user_1.discord]
        self.assertSetEqual(set(queryset), set(expected))



from unittest import TestCase

from ..exceptions import (
    DiscordApiBackoff, 
    DiscordClientException, 
    DiscordRateLimitExhausted,
    DiscordTooManyRequestsError
)


class TestExceptions(TestCase):

    def test_DiscordApiException(self):
        with self.assertRaises(DiscordClientException):
            raise DiscordClientException()

    def test_DiscordApiBackoff_raise(self):
        with self.assertRaises(DiscordApiBackoff):
            raise DiscordApiBackoff(999)

    def test_DiscordApiBackoff_retry_after_seconds(self):
        retry_after = 999
        ex = DiscordApiBackoff(retry_after)
        self.assertEqual(ex.retry_after, retry_after)
        self.assertEqual(ex.retry_after_seconds, 1)

    def test_DiscordRateLimitedExhausted_raise(self):
        with self.assertRaises(DiscordRateLimitExhausted):
            raise DiscordRateLimitExhausted(999)

    def test_DiscordApiBackoffError_raise(self):
        with self.assertRaises(DiscordTooManyRequestsError):
            raise DiscordTooManyRequestsError(999)



from unittest.mock import patch, Mock

from django.test import TestCase

from ..models import EveCharacter, EveCorporationInfo, EveAllianceInfo
from ..tasks import (
    update_alliance, 
    update_corp, 
    update_character, 
    run_model_update
)


class TestTasks(TestCase):

    @patch('allianceauth.eveonline.tasks.EveCorporationInfo')
    def test_update_corp(self, mock_EveCorporationInfo):
        update_corp(42)
        self.assertEqual(
            mock_EveCorporationInfo.objects.update_corporation.call_count, 1
        )
        self.assertEqual(
            mock_EveCorporationInfo.objects.update_corporation.call_args[0][0], 42
        )
            
    @patch('allianceauth.eveonline.tasks.EveAllianceInfo')
    def test_update_alliance(self, mock_EveAllianceInfo):
        update_alliance(42)
        self.assertEqual(
            mock_EveAllianceInfo.objects.update_alliance.call_args[0][0], 42
        )
        self.assertEqual(
            mock_EveAllianceInfo.objects
            .update_alliance.return_value.populate_alliance.call_count, 1
        )

    @patch('allianceauth.eveonline.tasks.EveCharacter')
    def test_update_character(self, mock_EveCharacter):
        update_character(42)
        self.assertEqual(
            mock_EveCharacter.objects.update_character.call_count, 1
        )
        self.assertEqual(
            mock_EveCharacter.objects.update_character.call_args[0][0], 42
        )


@patch('allianceauth.eveonline.tasks.update_character')
@patch('allianceauth.eveonline.tasks.update_alliance')
@patch('allianceauth.eveonline.tasks.update_corp')
@patch('allianceauth.eveonline.providers.provider')
@patch('allianceauth.eveonline.tasks.CHUNK_SIZE', 2)
class TestRunModelUpdate(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        EveCorporationInfo.objects.all().delete()
        EveAllianceInfo.objects.all().delete()
        EveCharacter.objects.all().delete()
        
        EveCorporationInfo.objects.create(
            corporation_id=2345,
            corporation_name='corp.name',
            corporation_ticker='c.c.t',
            member_count=10,
            alliance=None,
        )
        EveAllianceInfo.objects.create(
            alliance_id=3456,
            alliance_name='alliance.name',
            alliance_ticker='a.t',
            executor_corp_id=5,
        )       
        EveCharacter.objects.create(
            character_id=1,
            character_name='character.name1',
            corporation_id=2345,
            corporation_name='character.corp.name',
            corporation_ticker='c.c.t',  # max 5 chars 
            alliance_id=None
        )
        EveCharacter.objects.create(
            character_id=2,
            character_name='character.name2',
            corporation_id=9876,
            corporation_name='character.corp.name',
            corporation_ticker='c.c.t',  # max 5 chars 
            alliance_id=3456,
            alliance_name='character.alliance.name',
        )
        EveCharacter.objects.create(
            character_id=3,
            character_name='character.name3',
            corporation_id=9876,
            corporation_name='character.corp.name',
            corporation_ticker='c.c.t',  # max 5 chars 
            alliance_id=3456,
            alliance_name='character.alliance.name',
        )
        EveCharacter.objects.create(
            character_id=4,
            character_name='character.name4',
            corporation_id=9876,
            corporation_name='character.corp.name',
            corporation_ticker='c.c.t',  # max 5 chars 
            alliance_id=3456,
            alliance_name='character.alliance.name',
        )
        """
        EveCharacter.objects.create(
            character_id=5,
            character_name='character.name5',
            corporation_id=9876,
            corporation_name='character.corp.name',
            corporation_ticker='c.c.t',  # max 5 chars 
            alliance_id=3456,
            alliance_name='character.alliance.name',
        )
        """      
        
    def setUp(self):
        self.affiliations = [
            {'character_id': 1, 'corporation_id': 5},
            {'character_id': 2, 'corporation_id': 9876, 'alliance_id': 3456},
            {'character_id': 3, 'corporation_id': 9876, 'alliance_id': 7456},
            {'character_id': 4, 'corporation_id': 9876, 'alliance_id': 3456}
        ]
        self.names = [
            {'id': 1, 'name': 'character.name1'},
            {'id': 2, 'name': 'character.name2'},
            {'id': 3, 'name': 'character.name3'},
            {'id': 4, 'name': 'character.name4_new'}
        ]

    def test_normal_run(
        self,
        mock_provider,
        mock_update_corp,
        mock_update_alliance,
        mock_update_character,
    ):
        def get_affiliations(characters: list):
            response = [x for x in self.affiliations if x['character_id'] in characters]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        def get_names(ids: list):
            response = [x for x in self.names if x['id'] in ids]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        mock_provider.client.Character.post_characters_affiliation.side_effect \
            = get_affiliations
       
        mock_provider.client.Universe.post_universe_names.side_effect = get_names
        
        run_model_update()
        
        self.assertEqual(
            mock_provider.client.Character.post_characters_affiliation.call_count, 2
        )        
        self.assertEqual(
            mock_provider.client.Universe.post_universe_names.call_count, 2
        )

        # character 1 has changed corp
        # character 2 no change
        # character 3 has changed alliance
        # character 4 has changed name
        self.assertEqual(mock_update_corp.apply_async.call_count, 1)
        self.assertEqual(
            int(mock_update_corp.apply_async.call_args[1]['args'][0]), 2345
        )
        self.assertEqual(mock_update_alliance.apply_async.call_count, 1)
        self.assertEqual(
            int(mock_update_alliance.apply_async.call_args[1]['args'][0]), 3456
        )        
        characters_updated = {
            x[1]['args'][0] for x in mock_update_character.apply_async.call_args_list
        }
        excepted = {1, 3, 4}
        self.assertSetEqual(characters_updated, excepted)

    def test_ignore_character_not_in_affiliations(
        self,
        mock_provider,
        mock_update_corp,
        mock_update_alliance,
        mock_update_character,
    ):
        def get_affiliations(characters: list):
            response = [x for x in self.affiliations if x['character_id'] in characters]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        def get_names(ids: list):
            response = [x for x in self.names if x['id'] in ids]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        del self.affiliations[0]

        mock_provider.client.Character.post_characters_affiliation.side_effect \
            = get_affiliations
       
        mock_provider.client.Universe.post_universe_names.side_effect = get_names
        
        run_model_update()
        characters_updated = {
            x[1]['args'][0] for x in mock_update_character.apply_async.call_args_list
        }
        excepted = {3, 4}
        self.assertSetEqual(characters_updated, excepted)

    def test_ignore_character_not_in_names(
        self,
        mock_provider,
        mock_update_corp,
        mock_update_alliance,
        mock_update_character,
    ):
        def get_affiliations(characters: list):
            response = [x for x in self.affiliations if x['character_id'] in characters]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        def get_names(ids: list):
            response = [x for x in self.names if x['id'] in ids]
            mock_operator = Mock(**{'result.return_value': response})
            return mock_operator

        del self.names[3]

        mock_provider.client.Character.post_characters_affiliation.side_effect \
            = get_affiliations
       
        mock_provider.client.Universe.post_universe_names.side_effect = get_names
        
        run_model_update()
        characters_updated = {
            x[1]['args'][0] for x in mock_update_character.apply_async.call_args_list
        }
        excepted = {1, 3}
        self.assertSetEqual(characters_updated, excepted)



import logging

from celery import shared_task
from .models import EveAllianceInfo
from .models import EveCharacter
from .models import EveCorporationInfo

from . import providers

logger = logging.getLogger(__name__)

TASK_PRIORITY = 7
CHUNK_SIZE = 500


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


@shared_task
def update_corp(corp_id):
    """Update given corporation from ESI"""
    EveCorporationInfo.objects.update_corporation(corp_id)


@shared_task
def update_alliance(alliance_id):
    """Update given alliance from ESI"""
    EveAllianceInfo.objects.update_alliance(alliance_id).populate_alliance()


@shared_task
def update_character(character_id):
    """Update given character from ESI"""
    EveCharacter.objects.update_character(character_id)


@shared_task
def run_model_update():
    """Update all alliances, corporations and characters from ESI"""
    
    # update existing corp models
    for corp in EveCorporationInfo.objects.all().values('corporation_id'):
        update_corp.apply_async(
            args=[corp['corporation_id']], priority=TASK_PRIORITY
        )

    # update existing alliance models
    for alliance in EveAllianceInfo.objects.all().values('alliance_id'):
        update_alliance.apply_async(
            args=[alliance['alliance_id']], priority=TASK_PRIORITY
        )

    # update existing character models
    character_ids = EveCharacter.objects.all().values_list('character_id', flat=True)    
    for character_ids_chunk in chunks(character_ids, CHUNK_SIZE):
        affiliations_raw = providers.provider.client.Character\
            .post_characters_affiliation(characters=character_ids_chunk).result()
        character_names = providers.provider.client.Universe\
            .post_universe_names(ids=character_ids_chunk).result()

        affiliations = {
            affiliation.get('character_id'): affiliation 
            for affiliation in affiliations_raw
        }
        # add character names to affiliations
        for character in character_names:            
            character_id = character.get('id')
            if character_id in affiliations:
                affiliations[character_id]['name'] = character.get('name')

        # fetch current characters
        characters = EveCharacter.objects.filter(character_id__in=character_ids_chunk)\
            .values('character_id', 'corporation_id', 'alliance_id', 'character_name')  
        
        for character in characters:
            character_id = character.get('character_id')
            if character_id in affiliations:
                affiliation = affiliations[character_id]
                
                corp_changed = (
                    character.get('corporation_id') != affiliation.get('corporation_id')
                )
                
                alliance_id = character.get('alliance_id')
                if not alliance_id:
                    alliance_id = None
                alliance_changed = alliance_id != affiliation.get('alliance_id')
                
                name_changed = False
                fetched_name = affiliation.get('name', False)
                if fetched_name:
                    name_changed = character.get('character_name') != fetched_name
                
                if corp_changed or alliance_changed or name_changed:  
                    update_character.apply_async(
                        args=[character.get('character_id')], priority=TASK_PRIORITY
                    )



from django.shortcuts import render, get_object_or_404, redirect
from .models import Notification
from django.contrib.auth.decorators import login_required
from django.contrib import messages
from django.utils.translation import ugettext_lazy as _
import logging

logger = logging.getLogger(__name__)


@login_required
def notification_list(request):
    logger.debug("notification_list called by user %s" % request.user)
    new_notifs = Notification.objects.filter(user=request.user).filter(viewed=False)
    old_notifs = Notification.objects.filter(user=request.user).filter(viewed=True)
    logger.debug("User %s has %s unread and %s read notifications" % (request.user, len(new_notifs), len(old_notifs)))
    context = {
        'read': old_notifs,
        'unread': new_notifs,
    }
    return render(request, 'notifications/list.html', context)


@login_required
def notification_view(request, notif_id):
    logger.debug("notification_view called by user %s for notif_id %s" % (request.user, notif_id))
    notif = get_object_or_404(Notification, pk=notif_id)
    if notif.user == request.user:
        logger.debug("Providing notification for user %s" % request.user)
        context = {'notif': notif}
        notif.view()
        return render(request, 'notifications/view.html', context)
    else:
        logger.warn(
            "User %s not authorized to view notif_id %s belonging to user %s" % (request.user, notif_id, notif.user))
        messages.error(request, _('You are not authorized to view that notification.'))
        return redirect('notifications:list')


@login_required
def remove_notification(request, notif_id):
    logger.debug("remove notification called by user %s for notif_id %s" % (request.user, notif_id))
    notif = get_object_or_404(Notification, pk=notif_id)
    if notif.user == request.user:
        if Notification.objects.filter(id=notif_id).exists():
            notif.delete()
            logger.info("Deleting notif id %s by user %s" % (notif_id, request.user))
            messages.success(request, _('Deleted notification.'))
    else:
        logger.error(
            "Unable to delete notif id %s for user %s - notif matching id not found." % (notif_id, request.user))
        messages.error(request, _('Failed to locate notification.'))
    return redirect('notifications:list')


@login_required
def mark_all_read(request):
    logger.debug('mark all notifications read called by user %s' % request.user)
    Notification.objects.filter(user=request.user).update(viewed=True)
    messages.success(request, _('Marked all notifications as read.'))
    return redirect('notifications:list')


@login_required
def delete_all_read(request):
    logger.debug('delete all read notifications called by user %s' % request.user)
    Notification.objects.filter(user=request.user).filter(viewed=True).delete()
    messages.success(request, _('Deleted all read notifications.'))
    return redirect('notifications:list')



from unittest import mock

from django.contrib.auth.models import User, Group, Permission
from django.test import TestCase

from allianceauth.eveonline.models import (
    EveCorporationInfo, EveAllianceInfo, EveCharacter
)

from .auth_utils import AuthUtils


class TestAuthUtils(TestCase):

    def test_can_create_user(self):        
        user = AuthUtils.create_user('Bruce Wayne')
        self.assertTrue(User.objects.filter(username='Bruce Wayne').exists())
    
    def test_can_add_main_character_2(self):
        user = AuthUtils.create_user('Bruce Wayne')
        character = AuthUtils.add_main_character_2(
            user,
            name='Bruce Wayne',
            character_id=1001,
            corp_id=2001,
            corp_name='Wayne Technologies',
            corp_ticker='WYT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises'
        )
        expected = character
        self.assertEqual(user.profile.main_character, expected)
    
    def test_can_add_permission_to_group(self):
        group = Group.objects.create(name='Dummy Group')
        p = AuthUtils.get_permission_by_name('auth.group_management')
        AuthUtils.add_permissions_to_groups([p], [group])                
        self.assertTrue(group.permissions.filter(pk=p.pk).exists())

    def test_can_add_permission_to_user_by_name(self):
        user = AuthUtils.create_user('Bruce Wayne')
        user = AuthUtils.add_permission_to_user_by_name(
            'auth.timer_management', user
        )
        self.assertTrue(user.has_perm('auth.timer_management'))


class TestGetPermissionByName(TestCase):

    def test_can_get_permission_by_name(self):
        expected = Permission.objects.get(
            content_type__app_label='auth', codename='timer_management'
        )         
        self.assertEqual(
            AuthUtils.get_permission_by_name('auth.timer_management'), expected
        )

    def test_raises_exception_on_invalid_permission_format(self):
        with self.assertRaises(ValueError):
            AuthUtils.get_permission_by_name('timer_management')


from unittest.mock import patch

from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.test import TestCase, RequestFactory, Client

from allianceauth.authentication.models import CharacterOwnership, State
from allianceauth.eveonline.models import (
    EveCharacter, EveCorporationInfo, EveAllianceInfo
)

from ..admin import HasLeaderFilter, GroupAdmin, Group
from . import get_admin_change_view_url

if 'allianceauth.eveonline.autogroups' in settings.INSTALLED_APPS:
    _has_auto_groups = True
    from allianceauth.eveonline.autogroups.models import AutogroupsConfig
    from ..admin import IsAutoGroupFilter
else:
    _has_auto_groups = False


MODULE_PATH = 'allianceauth.groupmanagement.admin'


class MockRequest(object):
    
    def __init__(self, user=None):
        self.user = user


class TestGroupAdmin(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # group 1 - has leader
        cls.group_1 = Group.objects.create(name='Group 1')        
        cls.group_1.authgroup.description = 'Default Group'
        cls.group_1.authgroup.internal = False
        cls.group_1.authgroup.hidden = False
        cls.group_1.authgroup.save()

        # group 2 - no leader
        cls.group_2 = Group.objects.create(name='Group 2')
        cls.group_2.authgroup.description = 'Internal Group'
        cls.group_2.authgroup.internal = True        
        cls.group_2.authgroup.save()

        # group 3 - has leader
        cls.group_3 = Group.objects.create(name='Group 3')
        cls.group_3.authgroup.description = 'Hidden Group'
        cls.group_3.authgroup.internal = False
        cls.group_3.authgroup.hidden = True
        cls.group_3.authgroup.save()

        # group 4 - no leader
        cls.group_4 = Group.objects.create(name='Group 4')
        cls.group_4.authgroup.description = 'Open Group'
        cls.group_4.authgroup.internal = False
        cls.group_4.authgroup.hidden = False
        cls.group_4.authgroup.open = True
        cls.group_4.authgroup.save()

        # group 5 - no leader
        cls.group_5 = Group.objects.create(name='Group 5')
        cls.group_5.authgroup.description = 'Public Group'
        cls.group_5.authgroup.internal = False
        cls.group_5.authgroup.hidden = False
        cls.group_5.authgroup.public = True
        cls.group_5.authgroup.save()

        # group 6 - no leader
        cls.group_6 = Group.objects.create(name='Group 6')
        cls.group_6.authgroup.description = 'Mixed Group'
        cls.group_6.authgroup.internal = False
        cls.group_6.authgroup.hidden = True
        cls.group_6.authgroup.open = True
        cls.group_6.authgroup.public = True
        cls.group_6.authgroup.save()
        
        # user 1 - corp and alliance, normal user
        cls.character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name='Bruce Wayne',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        cls.character_1a = EveCharacter.objects.create(
            character_id=1002,
            character_name='Batman',
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',
        )
        alliance = EveAllianceInfo.objects.create(
            alliance_id=3001,
            alliance_name='Wayne Enterprises',
            alliance_ticker='WE',            
            executor_corp_id=2001
        )
        EveCorporationInfo.objects.create(
            corporation_id=2001,
            corporation_name='Wayne Technologies',
            corporation_ticker='WT',            
            member_count=42,
            alliance=alliance
        )        
        cls.user_1 = User.objects.create_user(
            cls.character_1.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_1,
            owner_hash='x1' + cls.character_1.character_name,
            user=cls.user_1
        )
        CharacterOwnership.objects.create(
            character=cls.character_1a,
            owner_hash='x1' + cls.character_1a.character_name,
            user=cls.user_1
        )
        cls.user_1.profile.main_character = cls.character_1
        cls.user_1.profile.save()
        cls.user_1.groups.add(cls.group_1)
        cls.group_1.authgroup.group_leaders.add(cls.user_1)

        # user 2 - corp only, staff
        cls.character_2 = EveCharacter.objects.create(
            character_id=1003,
            character_name='Clark Kent',
            corporation_id=2002,
            corporation_name='Daily Planet',
            corporation_ticker='DP',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2002,
            corporation_name='Daily Plane',
            corporation_ticker='DP',            
            member_count=99,
            alliance=None
        )
        cls.user_2 = User.objects.create_user(
            cls.character_2.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_2,
            owner_hash='x1' + cls.character_2.character_name,
            user=cls.user_2
        )
        cls.user_2.profile.main_character = cls.character_2
        cls.user_2.profile.save()
        cls.user_2.groups.add(cls.group_2)
        cls.user_2.is_staff = True
        cls.user_2.save()
        
        # user 3 - no main, no group, superuser
        cls.character_3 = EveCharacter.objects.create(
            character_id=1101,
            character_name='Lex Luthor',
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',
            alliance_id=None
        )
        EveCorporationInfo.objects.create(
            corporation_id=2101,
            corporation_name='Lex Corp',
            corporation_ticker='LC',            
            member_count=666,
            alliance=None
        )
        EveAllianceInfo.objects.create(
            alliance_id=3101,
            alliance_name='Lex World Domination',
            alliance_ticker='LWD',
            executor_corp_id=2101
        )
        cls.user_3 = User.objects.create_user(
            cls.character_3.character_name.replace(' ', '_'),
            'abc@example.com',
            'password'
        )
        CharacterOwnership.objects.create(
            character=cls.character_3,
            owner_hash='x1' + cls.character_3.character_name,
            user=cls.user_3
        )
        cls.user_3.is_superuser = True
        cls.user_3.save()
        cls.user_3.groups.add(cls.group_3)
        cls.group_3.authgroup.group_leaders.add(cls.user_3)

    def setUp(self):
        self.factory = RequestFactory()
        self.modeladmin = GroupAdmin(
            model=Group, admin_site=AdminSite()
        )

    def _create_autogroups(self):
        """create autogroups for corps and alliances"""
        if _has_auto_groups:
            autogroups_config = AutogroupsConfig(
                corp_groups=True,
                alliance_groups=True
            )
            autogroups_config.save()
            for state in State.objects.all():
                autogroups_config.states.add(state)
            autogroups_config.update_corp_group_membership(self.user_1)
    
    # column rendering

    def test_description(self):
        expected = 'Default Group'
        result = self.modeladmin._description(self.group_1)
        self.assertEqual(result, expected)

    def test_member_count(self):        
        expected = 1
        obj = self.modeladmin.get_queryset(MockRequest(user=self.user_1))\
            .get(pk=self.group_1.pk)
        result = self.modeladmin._member_count(obj)
        self.assertEqual(result, expected)

    def test_has_leader(self):        
        result = self.modeladmin.has_leader(self.group_1)
        self.assertTrue(result)

    def test_properties_1(self):
        expected = ['Default']
        result = self.modeladmin._properties(self.group_1)
        self.assertListEqual(result, expected)

    def test_properties_2(self):
        expected = ['Internal']
        result = self.modeladmin._properties(self.group_2)
        self.assertListEqual(result, expected)

    def test_properties_3(self):
        expected = ['Hidden']
        result = self.modeladmin._properties(self.group_3)
        self.assertListEqual(result, expected)

    def test_properties_4(self):
        expected = ['Open']
        result = self.modeladmin._properties(self.group_4)
        self.assertListEqual(result, expected)

    def test_properties_5(self):
        expected = ['Public']
        result = self.modeladmin._properties(self.group_5)
        self.assertListEqual(result, expected)

    def test_properties_6(self):
        expected = ['Hidden', 'Open', 'Public']
        result = self.modeladmin._properties(self.group_6)
        self.assertListEqual(result, expected)

    if _has_auto_groups:
        @patch(MODULE_PATH + '._has_auto_groups', True)
        def test_properties_7(self):
            self._create_autogroups()
            expected = ['Auto Group']
            my_group = Group.objects\
                .filter(managedcorpgroup__isnull=False)\
                .first()
            result = self.modeladmin._properties(my_group)
            self.assertListEqual(result, expected)

    # actions
   
    # filters
    
    if _has_auto_groups:
        @patch(MODULE_PATH + '._has_auto_groups', True)
        def test_filter_is_auto_group(self):
            
            class GroupAdminTest(admin.ModelAdmin): 
                list_filter = (IsAutoGroupFilter,)
                    
            self._create_autogroups()        
            my_modeladmin = GroupAdminTest(Group, AdminSite())

            # Make sure the lookups are correct
            request = self.factory.get('/')
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            filters = changelist.get_filters(request)
            filterspec = filters[0][0]
            expected = [
                ('yes', 'Yes'),
                ('no', 'No'),
            ]
            self.assertEqual(filterspec.lookup_choices, expected)

            # Make sure the correct queryset is returned - no
            request = self.factory.get(
                '/', {'is_auto_group__exact': 'no'}
            )
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            queryset = changelist.get_queryset(request)
            expected = [
                self.group_1,
                self.group_2,
                self.group_3,
                self.group_4,
                self.group_5,
                self.group_6
            ]
            self.assertSetEqual(set(queryset), set(expected))

            # Make sure the correct queryset is returned - yes
            request = self.factory.get(
                '/', {'is_auto_group__exact': 'yes'}
            )
            request.user = self.user_1
            changelist = my_modeladmin.get_changelist_instance(request)
            queryset = changelist.get_queryset(request)
            expected = Group.objects.exclude(
                managedalliancegroup__isnull=True, 
                managedcorpgroup__isnull=True
            )
            self.assertSetEqual(set(queryset), set(expected))
         
    def test_filter_has_leader(self):
        
        class GroupAdminTest(admin.ModelAdmin): 
            list_filter = (HasLeaderFilter,)
                
        self._create_autogroups()        
        my_modeladmin = GroupAdminTest(Group, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get('/')
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        filters = changelist.get_filters(request)
        filterspec = filters[0][0]
        expected = [
            ('yes', 'Yes'),
            ('no', 'No'),
        ]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned - no
        request = self.factory.get(
            '/', {'has_leader__exact': 'no'}
        )
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Group.objects.exclude(pk__in=[
            self.group_1.pk, self.group_3.pk
        ])
        self.assertSetEqual(set(queryset), set(expected))

        # Make sure the correct queryset is returned - yes
        request = self.factory.get(
            '/', {'has_leader__exact': 'yes'}
        )
        request.user = self.user_1
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = [
            self.group_1,            
            self.group_3
        ]
        self.assertSetEqual(set(queryset), set(expected))
    
    def test_change_view_loads_normally(self):
        User.objects.create_superuser(
            username='superuser', password='secret', email='admin@example.com'
        )
        c = Client()
        c.login(username='superuser', password='secret')                
        response = c.get(get_admin_change_view_url(self.group_1))
        self.assertEqual(response.status_code, 200)
